from ..resources import Texture, TextureMap
from ._base import Material
from ..utils import Color, assert_type


class BackgroundMaterial(Material):
    """Solid/Gradient background color.

    A background material that draws the background is a uniform color
    or in a gradient. The positional arguments are passed to ``set_colors()``.

    Parameters
    ----------
    colors : list
        A list of 0, 1, 2, or 4 colors to use in the background. If zero,
        defaults to monochrome black. If one, the background will be monochrome
        using the given color. If two, the background will be a gradient from
        bottom to top using the given colors. If four, the background will be a
        gradient with each corner set to a different color. The value at a given
        position is then based on the relative distance of that position to each
        corner.
    kwargs : Any
        Additional kwargs are passed to the base constructor
        (:class:`pygfx.materials.Material`).

    """

    uniform_type = dict(
        Material.uniform_type,
        color_bottom_left="4xf4",
        color_bottom_right="4xf4",
        color_top_left="4xf4",
        color_top_right="4xf4",
    )

    def __init__(
        self,
        *colors,
        alpha_mode="blend",
        depth_write=False,
        render_queue=1000,
        **kwargs,
    ):
        super().__init__(
            alpha_mode=alpha_mode,
            depth_write=depth_write,
            render_queue=render_queue,
            **kwargs,
        )
        self.set_colors(*colors)

    def set_colors(self, *colors):
        """Set the background colors. If one color is given, it will be used
        as a uniform color. If two colors are given, it will be used for
        the bottom and top. If four colors are given, it will be used for the
        four corners.
        """
        colors = [Color(c) for c in colors]
        if len(colors) == 0:
            self.color_bottom_left = (0, 0, 0, 1)
            self.color_bottom_right = (0, 0, 0, 1)
            self.color_top_left = (0, 0, 0, 1)
            self.color_top_right = (0, 0, 0, 1)
        elif len(colors) == 1:
            self.color_bottom_left = colors[0]
            self.color_bottom_right = colors[0]
            self.color_top_left = colors[0]
            self.color_top_right = colors[0]
        elif len(colors) == 2:
            self.color_bottom_left = colors[0]
            self.color_bottom_right = colors[0]
            self.color_top_left = colors[1]
            self.color_top_right = colors[1]
        elif len(colors) == 4:
            self.color_bottom_left = colors[0]
            self.color_bottom_right = colors[1]
            self.color_top_left = colors[2]
            self.color_top_right = colors[3]
        else:
            raise ValueError("Need 1, 2 or 4 colors.")

    @property
    def color_bottom_left(self):
        """The color in the bottom left corner."""
        return Color(self.uniform_buffer.data["color_bottom_left"])

    @color_bottom_left.setter
    def color_bottom_left(self, color):
        self.uniform_buffer.data["color_bottom_left"] = Color(color)
        self.uniform_buffer.update_full()

    @property
    def color_bottom_right(self):
        """The color in the bottom right corner."""
        return Color(self.uniform_buffer.data["color_bottom_right"])

    @color_bottom_right.setter
    def color_bottom_right(self, color):
        self.uniform_buffer.data["color_bottom_right"] = Color(color)
        self.uniform_buffer.update_full()

    @property
    def color_top_left(self):
        """The color in the top left corner."""
        return Color(self.uniform_buffer.data["color_top_left"])

    @color_top_left.setter
    def color_top_left(self, color):
        self.uniform_buffer.data["color_top_left"] = Color(color)
        self.uniform_buffer.update_full()

    @property
    def color_top_right(self):
        """The color in the top right corner."""
        return Color(self.uniform_buffer.data["color_top_right"])

    @color_top_right.setter
    def color_top_right(self, color):
        self.uniform_buffer.data["color_top_right"] = Color(color)
        self.uniform_buffer.update_full()


class BackgroundImageMaterial(BackgroundMaterial):
    """Image/Skybox background.

    A background material that displays an image. If map is a 2D
    texture view, it is used as a static background. If it is a cube
    texture view, (on a NxMx6 texture) it is used as a skybox.

    Parameters
    ----------
    map : Texture
        If map is a 2D texture, it is used as static background image. If map is
        a cube texture, it is used as a skybox.
    kwargs : Any
        Additional kwargs are passed to the base constructor
        (:class:`pygfx.materials.Material`).

    """

    def __init__(self, map=None, **kwargs):
        super().__init__(**kwargs)
        self.map = map

    @property
    def map(self):
        """The texture map specifying the background image"""
        return self._store.map

    @map.setter
    def map(self, map):
        assert_type("map", map, None, Texture, TextureMap)
        if isinstance(map, Texture):
            map = TextureMap(map)
        self._store.map = map


class BackgroundSkyboxMaterial(BackgroundImageMaterial):
    """Skybox background.

    A cube image background, resulting in a skybox.

    """

    def __init__(self, map=None, **kwargs):
        super().__init__(map=map, **kwargs)



"""
This subpackage defines the base shader class. Basically this is where the wgsl
code is generated, composed, mixed and juggled.


## A note about pygfx shaders

In every render engine that uses (GPU) shaders, one particularly nasty problem
is that of shader composition. I.e. how to write shaders and shader snippets,
such that duplicate code is avoided, shaders can be produced for a wide variety
of parameters (i.e. materials and their properties), and still keep the system
understandable.

The way we've solved this in pygfx is by no means a holy grail, but it does feel
(to us) like a local optimum that's pretty flexible while keeping things simple.

Basically, based on the class of the world object and material, the
corresponding shader class is selected. Shader classes produce templated wgsl
(jinja2), which is then transformed into the final shader code using templating
variables, based on properties of the material and geometry, as well as external
features such as the render pass and renderstate (e.g. lights).

Wgsl templates can be provided in-line or loaded from wgsl files. Jinja2's
include mechanics is used to write snippets that can be used in multiple
places.
"""

from .base import ShaderInterface, BaseShader  # noqa
from .templating import register_wgsl_loader  # noqa



from ._base import Material
from ..utils import unpack_bitfield, Color


class TextMaterial(Material):
    """Basic text material.

    Parameters
    ----------
    color : Color
        The color of the text.
    outline_color : Color
        The color of the outline of the text.
    outline_thickness : int
        A value indicating the relative width of the outline. Valid values are
        between 0.0 and 0.5.
    weight_offset : int
        A value representing an offset to the font weight. Font weights are in
        the range 100-900, so this value should be in the same order of
        magnitude. Can be negative to make text thinner. Default zero.
    aa : bool
        Whether the glyphs is anti-aliased in the shader. Default False.
    kwargs : Any
        Additional kwargs will be passed to the :class:`material base class
        <pygfx.Material>`.

    Notes
    -----
    One use-case for weight_offset is to make dark text on a light background 50
    units thicker, to counter the psychological effect that such text *looks*
    thinner than a light text on a dark background.

    """

    uniform_type = dict(
        Material.uniform_type,
        color="4xf4",
        weight_offset="f4",
        outline_thickness="f4",
        outline_color="4xf4",
    )

    def __init__(
        self,
        color="#fff",
        *,
        outline_color="#000",
        outline_thickness=0,
        weight_offset=0,
        aa=False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.color = color
        self.outline_color = outline_color
        self.outline_thickness = outline_thickness
        self.weight_offset = weight_offset
        self.aa = aa

    def _wgpu_get_pick_info(self, pick_value):
        # Note that the glyph index is not necessarily the same as the
        # char index. It would not be worth the effort to let the shader produce
        # the char index, but I think we could make it possible to look
        # up the char index from the glyph index via the geometry.
        values = unpack_bitfield(pick_value, wobject_id=20, index=26, x=9, y=9)
        return {
            "glyph_index": values["index"],
            "point_coord": (values["x"] / 511.0, values["y"] / 511.0),
        }

    @property
    def aa(self):
        """Whether the glyphs edges are anti-aliased.

        Aliasing gives prettier results by producing semi-transparent fragments
        at the edges. Lines thinner than one physical pixel are also diminished
        by making them more transparent.

        However, because semi-transparent fragments are introduced, artifacts
        may occur if certain cases. For the same reason, aa only works for the
        "blended" and "weighted" alpha methods.

        Note that by default, pygfx already uses SSAA and/or PPAA to anti-alias
        the total renderered result. Text-based aa is an *additional* visual
        improvement.
        """
        return self._store.aa

    @aa.setter
    def aa(self, aa):
        self._store.aa = bool(aa)

    @property
    def _gfx_effective_aa(self):
        aa_able_methods = ("blended", "weighted")
        return self._store.aa and self.alpha_method in aa_able_methods

    @property
    def color(self):
        """The color of the text."""
        return Color(self.uniform_buffer.data["color"])

    @color.setter
    def color(self, color):
        color = Color(color)
        self.uniform_buffer.data["color"] = color
        self.uniform_buffer.update_full()

    @property
    def outline_thickness(self):
        """A value indicating the relative width of the outline. Valid
        values are between 0.0 and 0.5. Default 0 (no outline).
        """
        return float(self.uniform_buffer.data["outline_thickness"])

    @outline_thickness.setter
    def outline_thickness(self, value):
        self.uniform_buffer.data["outline_thickness"] = max(0.0, min(0.5, float(value)))
        self.uniform_buffer.update_full()

    @property
    def outline_color(self):
        """The color of the outline of the text."""
        return Color(self.uniform_buffer.data["outline_color"])

    @outline_color.setter
    def outline_color(self, outline_color):
        outline_color = Color(outline_color)
        self.uniform_buffer.data["outline_color"] = outline_color
        self.uniform_buffer.update_full()

    @property
    def weight_offset(self):
        """A value representing an offset to the font weight. Font weights
        are in the range 100-900, so this value should be in the same order of
        magnitude. Can be negative to make text thinner. Default zero.

        One use-case is to make dark text on a light background 50 units
        thicker, to counter the psychological effect that such text
        *looks* thinner than a light text on a dark background.
        """
        return float(self.uniform_buffer.data["weight_offset"])

    @weight_offset.setter
    def weight_offset(self, value):
        self.uniform_buffer.data["weight_offset"] = float(value)
        self.uniform_buffer.update_full()



import math

import pylinalg as la
import numpy as np

from ._base import WorldObject
from ..utils.color import Color
from ..cameras import Camera
from ..resources import Buffer
from ..cameras import PerspectiveCamera
from ..utils import array_from_shadertype


def get_pos_from_camera_parent_or_target(light: "Light") -> np.ndarray:
    if isinstance(light.parent, Camera):
        cam = light.parent
        transform = cam.world.matrix
        return la.vec_transform((0, 0, -1), transform)
    elif isinstance(light, SpotLight):
        return light.target.world.position
    elif isinstance(light, DirectionalLight):
        return light.target.world.position
    else:
        raise ValueError("Unknown light source.")


class Light(WorldObject):
    """Light Base Class.

    Parameters
    ----------
    color : Color
        The color of the light emitted.
    intensity : float
        The light's intensity. Its units depend on the type of light. For point
        and spot lights it represents the luminous intensity of the light
        measured in candela (cd).
    cast_shadow : bool
        If True, the light can cast shadows. Otherwise it doesn't.

    Notes
    -----
    The light's intensity scales the color in the physical colorspace, as if
    scaling the number of photons. Note that an intensity of 0.5 is not
    equivalent to halving the color value. This is because the srgb color is
    perceptually linear, while intensity is physically linear. Values over 1.0
    make perfect sense - it's just a brighter light.

    There are two booleans that control the behavior of shadow: `cast_shadow` on
    a Light, and `receive_shadow` on the illuminated object. Shadows will only
    be displayed if both are set to True.

    """

    # Note that for lights and shadows, the uniform data is stored on the renderstate.
    # We can use the uniform_buffer as usual though. We'll just copy it over.

    _FORWARD_IS_MINUS_Z = True

    uniform_type = dict(
        WorldObject.uniform_type,
        color="4xf4",
        intensity="f4",
        cast_shadow="i4",
        light_view_proj_matrix="4x4xf4",
        shadow_bias="f4",
    )

    def __init__(self, color="#ffffff", intensity=1, *, cast_shadow=False, **kwargs):
        super().__init__(**kwargs)
        self.color = color
        self.intensity = intensity
        self.cast_shadow = cast_shadow

        # for internal use
        self._shadow = None

    def _gfx_update_uniform_buffer(self):
        # _gfx prefix means that its not a public method, but other parts in pygfx can use it.
        pass

    @property
    def shadow(self):
        """The shadow object for this light."""
        return self._shadow

    @property
    def color(self):
        """The color of the light, in the srgb colorspace."""
        return Color(self.uniform_buffer.data["color"])

    @color.setter
    def color(self, color):
        self.uniform_buffer.data["color"] = Color(color)
        self.uniform_buffer.update_full()

    @property
    def intensity(self):
        """The light intensity as a float, default 1.0. The units of
        intensity depend on the type of light. For point and spot lights
        it represents the luminous intensity of the light measured in
        candela (cd).

        The intensity scales the color in the physical colorspace, as
        if scaling the number of photons. Note that an intensity of 0.5
        is not equivalent to halving the color value. This is because
        the srgb color is perceptually linear, while intensity is
        physically linear. Values over 1.0 make perfect sense - it's
        just a brighter light.
        """
        return float(self.uniform_buffer.data["intensity"])

    @intensity.setter
    def intensity(self, value):
        self.uniform_buffer.data["intensity"] = float(value)
        self.uniform_buffer.update_full()

    @property
    def cast_shadow(self):
        """Whether or not this light will casts shadows on objects.
        Note that shadows are only cast on objects that have receive_shadow
        set to True.
        """
        return bool(self.uniform_buffer.data["cast_shadow"])

    @cast_shadow.setter
    def cast_shadow(self, value: bool):
        self.uniform_buffer.data["cast_shadow"] = bool(value)


class AmbientLight(Light):
    """Ambient light source.

    A light that omnidirectionally illuminates all objects in the scene equally.

    Parameters
    ----------
    color : Color
        The color of the emitted light.
    intensity : float
        The light intensity. A value of ``0.2`` corresponds to a dimly lit
        scene.

    """

    def __init__(self, color="#ffffff", intensity=0.2):
        super().__init__(color, intensity)


class PointLight(Light):
    """Radial point light source.

    A light that gets emitted from a single point in all directions.

    Parameters
    ----------
    color : Color
        The color of the emitted light.
    intensity : float
        The light intensity. A value of ``3`` corresponds to a well lit
        scene.
    cast_shadow : bool
        If True, the light can cast shadows. Otherwise it doesn't.
    distance : float
        The maximum distance at which objects are considered illuminated by the
        light. Limiting this may increase performance in large scenes. A value
        of ``0`` means that all objects are considered.
    decay : float
        The rate at which the light dims as it travels. A value of ``0`` means
        no decay. A decay of ``2`` is physically correct.

    Notes
    -----
    When setting ``decay`` to a non-zero value you likely probably have to
    increase intensity (a lot) as light decays following an inverse-square
    profile.

    """

    uniform_type = dict(
        Light.uniform_type,
        distance="f4",
        decay="f4",
        light_view_proj_matrix="6*4x4xf4",
    )

    def __init__(
        self,
        color="#ffffff",
        intensity=3,
        *,
        cast_shadow=False,
        distance=0,
        decay=0,
        **kwargs,
    ):
        super().__init__(color, intensity, cast_shadow=cast_shadow, **kwargs)
        self.distance = distance
        self.decay = decay
        self._shadow = PointLightShadow()

    @property
    def power(self):
        """The light's power. I.e. the luminous power of the light measured in lumens (lm).
        Changing the power will also change the light's intensity.
        """
        # compute the light's luminous power (in lumens) from its intensity (in candela)
        # for an isotropic light source, luminous power (lm) = 4 π luminous intensity (cd)
        return self.intensity * 4 * np.pi

    @power.setter
    def power(self, power):
        # set the light's intensity (in candela) from the desired luminous power (in lumens)
        self.intensity = power / (4 * np.pi)

    @property
    def distance(self):
        """When distance is zero, light will attenuate according to
        inverse-square law to infinite distance. When distance is
        non-zero, light will attenuate the same way until near the
        distance cutoff, where it will then attenuate quickly and
        smoothly to 0. Inherently, cutoffs are not physically correct.
        """
        return float(self.uniform_buffer.data["distance"])

    @distance.setter
    def distance(self, value):
        self.uniform_buffer.data["distance"] = value
        self.uniform_buffer.update_full()

    @property
    def decay(self):
        """The amount the light dims along the distance of the light.
        A decay of 2 leads to physically realistic light falloff.
        Default is 0, which means the light does not decay.
        """
        return float(self.uniform_buffer.data["decay"])

    @decay.setter
    def decay(self, value):
        self.uniform_buffer.data["decay"] = value
        self.uniform_buffer.update_full()


class DirectionalLight(Light):
    """Directional light source.

    A light that gets emitted in a direction, specified by its position and a
    target. This is equivalent to an infinitely large softbox.

    Parameters
    ----------
    color : Color
        The color of the light emitted.
    intensity : float
        The light intensity. A value of ``3`` corresponds to a well lit
        scene.
    cast_shadow : bool
        If True, the light can cast shadows. Otherwise it doesn't.
    target : WorldObject
        The object used to determine the light's direction. The light will shine
        from it's position toward's the direction of the target except when the
        light's parent is a camera, in which case target is ignored.

    Notes
    -----
    If this light is attached to a camera it's direction will follow the camera
    view direction.

    There are two booleans that control the behavior of shadow: `cast_shadow` on
    a Light, and `receive_shadow` on the illuminated object. Shadows will only
    be displayed if both are set to True.

    """

    uniform_type = dict(
        Light.uniform_type,
        direction="4xf4",
    )

    def __init__(
        self, color="#ffffff", intensity=3, *, cast_shadow=False, target=None, **kwargs
    ):
        super().__init__(color, intensity, cast_shadow=cast_shadow, **kwargs)
        self.target = target or WorldObject()
        self._shadow = DirectionalLightShadow()
        self._gfx_distance_to_target = 0

    @property
    def target(self):
        """The object to direct the light at. The light direction is
        from its position to its target.

        However, if the light's parent is a camera, it follows the
        camera direction instead (thus ignoring the target).
        """
        return self._target

    @target.setter
    def target(self, target):
        assert isinstance(target, WorldObject)
        self._target = target

    def _gfx_update_uniform_buffer(self):
        pos1 = self.world.position
        pos2 = get_pos_from_camera_parent_or_target(self)
        origin_to_target = pos2 - pos1
        self._gfx_distance_to_target = np.linalg.norm(origin_to_target)
        if self._gfx_distance_to_target > 0:
            direction = origin_to_target / self._gfx_distance_to_target
        else:
            direction = np.array(
                (0, 0, -1), dtype=float
            )  # ill-defined direction -> look neg z-axis
        self.uniform_buffer.data["direction"].flat = direction
        self.look_at(pos1 + direction)


class SpotLight(Light):
    """Directional point light source.

    A light that gets emitted from a single point in one direction, along a cone
    that increases in size the further from the light it gets.

    Parameters
    ----------
    color : Color
        The color of the light emitted.
    intensity : float
        The light intensity. A value of ``3`` corresponds to a well lit
        scene.
    cast_shadow : bool
        If True, the light can cast shadows. Otherwise it doesn't.
    distance : float
        The maximum distance at which objects are considered illuminated by the
        light. Limiting this may increase performance in large scenes. A value
        of ``0`` means that all objects are considered.
    decay : float
        The rate at which the light dims as it travels. A value of ``0`` means
        no decay. A decay of ``2`` is physically correct.
    angle : float
        The central angle (in rad) of the light's cone.
    penumbra : float
        Percent of the spotlight cone that is attenuated due
        to penumbra. Takes values between zero and 1. Default is zero.

    Notes
    -----
    If this light is attached to a camera it's direction will follow the camera
    view direction.

    There are two booleans that control the behavior of shadow: `cast_shadow` on
    a Light, and `receive_shadow` on the illuminated object. Shadows will only
    be displayed if both are set to True.

    When setting ``decay`` to a non-zero value you likely probably have to
    increase intensity (a lot) as light decays following an inverse-square
    profile.

    """

    uniform_type = dict(
        Light.uniform_type,
        direction="4xf4",
        distance="f4",
        cone_cos="f4",
        penumbra_cos="f4",
        decay="f4",
    )

    def __init__(
        self,
        color="#ffffff",
        intensity=3,
        *,
        cast_shadow=False,
        distance=0,
        decay=0,
        angle=math.pi / 3,
        penumbra=0,
        **kwargs,
    ):
        super().__init__(color, intensity, cast_shadow=cast_shadow, **kwargs)

        self.distance = distance
        self._angle = angle
        self._penumbra = penumbra
        self.decay = decay
        self.target = WorldObject()

        self.angle = angle
        self.penumbra = penumbra

        self._shadow = SpotLightShadow()

    def _gfx_update_uniform_buffer(self):
        pos1 = self.world.position
        pos2 = get_pos_from_camera_parent_or_target(self)
        origin_to_target = pos2 - pos1
        self._gfx_distance_to_target = np.linalg.norm(origin_to_target)
        if self._gfx_distance_to_target > 0:
            direction = origin_to_target / self._gfx_distance_to_target
        else:
            direction = np.array(
                (0, 0, -1), dtype=float
            )  # ill-defined direction -> look neg z-axis
        self.uniform_buffer.data["direction"].flat = direction
        self.look_at(pos1 + direction)

    @property
    def power(self):
        """The light's power. I.e. the luminous power of the light measured in lumens (lm).
        Changing the power will also change the light's intensity.
        """
        # compute the light's luminous power (in lumens) from its intensity (in candela)
        # by convention for a spotlight, luminous power (lm) = π * luminous intensity (cd)
        return self.intensity * np.pi

    @power.setter
    def power(self, power):
        # set the light's intensity (in candela) from the desired luminous power (in lumens)
        self.intensity = power / np.pi

    @property
    def distance(self):
        """When distance is zero, light will attenuate according to
        inverse-square law to infinite distance. When distance is
        non-zero, light will attenuate the same way until near the
        distance cutoff, where it will then attenuate quickly and
        smoothly to 0. Inherently, cutoffs are not physically correct.
        """
        return float(self.uniform_buffer.data["distance"])

    @distance.setter
    def distance(self, value):
        self.uniform_buffer.data["distance"] = value
        self.uniform_buffer.update_full()

    @property
    def angle(self):
        """The maximum extent of the spotlight, in radians, from its
        direction. Should be no more than Math.PI/2.
        """
        return self._angle

    @angle.setter
    def angle(self, value):
        self._angle = value
        cone_cos = math.cos(self._angle)
        self.uniform_buffer.data["cone_cos"] = cone_cos
        penumbra_cos = math.cos(self.angle * (1 - self.penumbra))
        self.uniform_buffer.data["penumbra_cos"] = penumbra_cos
        self.uniform_buffer.update_full()

    @property
    def penumbra(self):
        """Percent of the spotlight cone that is attenuated due to penumbra.
        Takes values between zero and 1.
        """
        return self._penumbra

    @penumbra.setter
    def penumbra(self, value):
        self._penumbra = value
        penumbra_cos = math.cos(self.angle * (1 - self.penumbra))
        self.uniform_buffer.data["penumbra_cos"] = penumbra_cos
        self.uniform_buffer.update_full()

    @property
    def decay(self):
        """The amount the light dims along the distance of the light.
        A decay of 2 leads to physically realistic light falloff.
        Default is 0, which means the light does not decay.
        """
        return float(self.uniform_buffer.data["decay"])

    @decay.setter
    def decay(self, value):
        self.uniform_buffer.data["decay"] = value
        self.uniform_buffer.update_full()


# shadows
shadow_uniform_type = dict(light_view_proj_matrix="4x4xf4")


class LightShadow:
    """Shadow map utility base class.

    This is usually created automatically by a light, and can be accessed
    through the light's shadow property.

    Parameters
    ----------
    camera : Camera
        The light's view of the scene. This is used to generate a depth map of
        the scene; objects occluded by other objects from the light's
        perspective will receive shadow.

    """

    def __init__(self, camera: PerspectiveCamera) -> None:
        self._camera = camera
        self._camera.maintain_aspect = False

        # TODO: 'radius' represents the shadow sampling radius,
        # which is used for PCF to blur and smooth shadow edges.
        # But it seems difficult to be used as a uniform in shader internal.
        # Changing this value will cause the shader to recompile.
        # Shadows with different radius in one scene are also difficult to handle.
        # now, it is a fixed value in shader.
        # self._radius = 1
        # self._map_size = [1024, 1024]

        # TODO: move bias and cull_mode to Light so they can be reactive?
        self.bias = 0
        self.cull_mode = "front"
        self._gfx_matrix_buffer = Buffer(
            array_from_shadertype(shadow_uniform_type), force_contiguous=True
        )

    @property
    def camera(self):
        """The camera that defines the POV for determining the depth map of the scene."""
        return self._camera

    @property
    def bias(self):
        """Shadow map bias. Very tiny adjustments here may help reduce artifacts in shadows."""
        return self._bias

    @bias.setter
    def bias(self, value):
        self._bias = float(value)

    @property
    def cull_mode(self):
        """
        Shadow map cull_mode ('front', 'back', or 'none'). When shadow
        mapping open meshes, set to 'none' and increase ``bias`` value
        to avoid shadow acne.
        """
        return self._cull_mode

    @cull_mode.setter
    def cull_mode(self, value):
        value = str(value).lower()
        if value in ("front", "back", "none"):
            self._cull_mode = value
        else:
            raise ValueError(f"invalid cull_mode: '{value}'")

    def _gfx_update_uniform_buffer(self, light: Light):
        light.uniform_buffer.data["shadow_bias"] = self._bias
        self._update_matrix(light)
        light.uniform_buffer.update_full()

    def _update_matrix(self, light: Light) -> None:
        shadow_camera = self.camera
        shadow_camera.world.position = light.world.position
        target = get_pos_from_camera_parent_or_target(light)
        shadow_camera.look_at(target)

        self._gfx_matrix_buffer.data["light_view_proj_matrix"] = (
            shadow_camera.camera_matrix.T
        )
        self._gfx_matrix_buffer.update_full()

        m = shadow_camera.camera_matrix.T
        light.uniform_buffer.data["light_view_proj_matrix"] = m


class DirectionalLightShadow(LightShadow):
    """Shadow map utility for directional lights."""

    def __init__(self) -> None:
        # orhto projection (fov==0) for directional light
        super().__init__(
            PerspectiveCamera(0, width=1000, height=1000, depth_range=(-500, 500))
        )

    def _update_matrix(self, light):
        super()._update_matrix(light)


class SpotLightShadow(LightShadow):
    """Shadow map utility for spot light sources."""

    def __init__(self) -> None:
        super().__init__(PerspectiveCamera(50, depth_range=(0.5, 500)))
        self._focus = 1

    def _update_matrix(self, light):
        camera = self.camera

        fov = 180 / math.pi * 2 * light.angle * self._focus

        aspect = 1
        far = (light.distance * 10) or camera.far

        if fov != camera.fov or far != camera.far:
            camera.fov = fov
            camera.aspect = aspect
            camera.depth_range = far / 1000000, far

        super()._update_matrix(light)


class PointLightShadow(LightShadow):
    """Shadow map utility for point light sources."""

    _cube_directions = np.array(
        [
            (1, 0, 0),
            (-1, 0, 0),
            (0, 1, 0),
            (0, -1, 0),
            (0, 0, 1),
            (0, 0, -1),
        ],
        dtype=float,
    )

    def __init__(self) -> None:
        # Note, this initially did not set depth_range. The depth_range is now
        # set explicitly to the range the camera calculated automatically, so
        # that the validation examples don't change. It would probably be best
        # to set all depth_ranges consistently, and based on the size of the scene.
        super().__init__(PerspectiveCamera(90, depth_range=(0.5, 100000)))

        self._gfx_matrix_buffer = []

        for _ in range(6):
            buffer = Buffer(
                array_from_shadertype(shadow_uniform_type), force_contiguous=True
            )
            self._gfx_matrix_buffer.append(buffer)

    def _update_matrix(self, light: Light) -> None:
        camera = self.camera
        camera.world.position = light.world.position
        directions = self._cube_directions + light.world.position

        far = (light.distance * 10) or camera.far

        if far != camera.far:
            camera.depth_range = far / 1000000, far

        for i in range(6):
            # Note: the direction may align with `up`, but we have logic in
            # `look_at` to catch and handle this special case.
            camera.look_at(directions[i])

            m = camera.camera_matrix.T
            light.uniform_buffer.data["light_view_proj_matrix"][i] = m
            self._gfx_matrix_buffer[i].data["light_view_proj_matrix"] = m
            self._gfx_matrix_buffer[i].update_full()
        light.uniform_buffer.update_full()



"""
This module implements the PipelineContainer (for compute and render pipelines).
This object is responsible for creating the native wgpu objects and doing the
actual dispatching / drawing.
"""

import sys
import wgpu

import os

from ....utils import logger
from ....utils.weak import WeakAssociativeContainer
from ....utils.trackable import PropTracker

from ..shader.base import ShaderInterface
from .utils import registry, GpuCache, hash_from_value
from .shared import get_shared
from .binding import Binding


PIPELINE_CONTAINER_GROUPS = WeakAssociativeContainer()

# These caches use an LRU strategy.
# They store a limited number of recently used GPU resources to improve cache hit rates.
# This enables reuse of GPU resources across similar objects, reducing the overhead of
# repeatedly creating/destroying GPU resources, especially in dynamic rendering scenarios.
LAYOUT_CACHE = GpuCache("layouts")
BINDING_CACHE = GpuCache("bindings")
SHADER_CACHE = GpuCache("shader_modules")
PIPELINE_CACHE = GpuCache("pipelines")

PRINT_WGSL_ON_ERROR = os.environ.get(
    "PYGFX_PRINT_WGSL_ON_COMPILATION_ERROR", "0"
).lower() not in ["false", "0"]


def get_cached_bind_group_layout(device, *args):
    key = "bind_group_layout", hash_from_value(args)
    result = LAYOUT_CACHE.get(key)
    if result is None:
        (entries,) = args
        result = device.create_bind_group_layout(entries=entries)

        LAYOUT_CACHE.set(key, result)

    return result


def get_cached_pipeline_layout(device, *args):
    key = "pipeline_layout", hash_from_value(args)
    result = LAYOUT_CACHE.get(key)
    if result is None:
        (bind_group_layouts,) = args
        result = device.create_pipeline_layout(bind_group_layouts=bind_group_layouts)

        LAYOUT_CACHE.set(key, result)

    return result


def get_cached_bind_group(device, *args):
    key = "bind_group", hash_from_value(args)
    result = BINDING_CACHE.get(key)
    if result is None:
        layout, entries = args
        result = device.create_bind_group(layout=layout, entries=entries)

        BINDING_CACHE.set(key, result)

    return result


def get_cached_shader_module(device, shader, shader_kwargs):
    # Using a key that *defines* the wgsl - rather than the wgsl itself
    # - avoids the templating to be applied on a cache hit, which saves
    # considerable time!
    key = "shader", shader.hash, hash_from_value(shader_kwargs)

    result = SHADER_CACHE.get(key)
    if result is None:
        wgsl = shader.generate_wgsl(**shader_kwargs)
        try:
            result = device.create_shader_module(code=wgsl)
        except wgpu.GPUValidationError:
            # No need to be super fancy in the formatting but we want to
            # help the users find their bugs
            # We have seen some shaders with close to 1000 lines of code
            # So we print numbers that would be aligned up to 5 digits.
            # Since this error may be confusing for end users due to the sheer
            # volume of text printed, developers must enable this with
            # PYGFX_PRINT_WGSL_ON_COMPILATION_ERROR
            if PRINT_WGSL_ON_ERROR:
                wgsl_with_line_numbers = "\n".join(
                    f"{i + 1:5d}: {line}" for i, line in enumerate(wgsl.splitlines())
                )
                print(wgsl_with_line_numbers, file=sys.stderr)
            raise

        SHADER_CACHE.set(key, result)

    return result


def get_cached_compute_pipeline(device, *args):
    key = "compute_pipeline", hash_from_value(args)
    result = PIPELINE_CACHE.get(key)
    if result is None:
        pipeline_layout, shader_module = args

        result = device.create_compute_pipeline(
            layout=pipeline_layout,
            compute={"module": shader_module, "entry_point": "main"},
        )

        PIPELINE_CACHE.set(key, result)

    return result


def get_cached_render_pipeline(device, *args):
    key = "render_pipeline", hash_from_value(args)
    result = PIPELINE_CACHE.get(key)
    if result is None:
        (
            pipeline_layout,
            shader_module,
            primitive_topology,
            strip_index_format,
            cull_mode,
            depth_descriptor,
            color_descriptors,
        ) = args

        result = device.create_render_pipeline(
            layout=pipeline_layout,
            vertex={
                "module": shader_module,
                "entry_point": "vs_main",
                "buffers": [],
            },
            primitive={
                "topology": primitive_topology,
                "strip_index_format": strip_index_format,
                "front_face": wgpu.FrontFace.ccw,
                "cull_mode": cull_mode,
            },
            depth_stencil=depth_descriptor,
            multisample={
                "count": 1,
                "mask": 0xFFFFFFFF,
                "alpha_to_coverage_enabled": False,
            },
            fragment={
                "module": shader_module,
                "entry_point": "fs_main",
                "targets": color_descriptors,
            },
        )

        PIPELINE_CACHE.set(key, result)

    return result


def get_pipeline_container_group(wobject, renderstate):
    """Return the PipelineContainerGroup object for wobject.

    This is the entrypoint for the renderer.

    The returned object has attributes ``bake_functions``, ``compute_containers``, and ``render_containers``.
    The latter two are lists to PipelineContainer objects. Most wobjects have just a single pipeline,
    but some may do preprocessing in a compute shader and/or consist of multiple passes.

    This call updates the GPU objects associated with the given wobject. It returns
    quickly if no changes are needed.
    """

    # Get pipeline container group. They are associated weakly by wobject and
    # renderstate. By also associating with the material the material can be
    # used by users to 'store' the pipeline, and hot-swap it. When any of these
    # objects (wobject, renderstate, material) is removed by the gc, the
    # associated pipeline_container object is removed as well.
    pcg_key = wobject, renderstate, wobject.material

    pipeline_container_group = PIPELINE_CONTAINER_GROUPS.get(pcg_key, None)
    if pipeline_container_group is None:
        pipeline_container_group = PipelineContainerGroup()
        PIPELINE_CONTAINER_GROUPS[pcg_key] = pipeline_container_group

    # Update. Quickly returns if no work to do.
    pipeline_container_group.update(wobject, renderstate)

    # Return the pipeline container group
    return pipeline_container_group


class PipelineContainerGroup:
    """Pipeline countainer group.

    This is a thin wrapper for a list of compute pipeline containers and render pipeline containers.
    The purpose of this object is to obtain the appropriate shader objects and store them.
    """

    def __init__(self):
        self.tracker = PropTracker()
        self.initialized = False
        # Public attributes used by the renderer
        self.compute_containers = None
        self.render_containers = None
        self.bake_functions = None

    def update(self, wobject, renderstate):
        """Update the pipeline containers. Creates (and re-creates) the containers if necessary."""

        # Get what has changed
        tracker = self.tracker
        changed = tracker.pop_changed()

        if not self.initialized:
            self.initialized = True
            changed.add("create")

        if not changed:
            return

        if "create" in changed:
            self.compute_containers = ()
            self.render_containers = ()
            self.bake_functions = ()

            # Get render function for this world object,
            # and use it to get a high-level description of pipelines.
            renderfunc = registry.get_render_function(wobject)
            if renderfunc is None:
                raise ValueError(
                    f"Could not get a render function for {wobject.__class__.__name__} "
                    f"with {wobject.material.__class__.__name__}"
                )

            # Call render function
            with tracker.track_usage("create"):
                shaders = renderfunc(wobject)
                if isinstance(shaders, ShaderInterface):
                    shaders = [shaders]

            # Divide result over two bins, one for compute, and one for render. Plus collect bake funcs.
            compute_containers = []
            render_containers = []
            bake_functions = []
            for shader in shaders:
                assert isinstance(shader, ShaderInterface)
                if shader.type == "compute":
                    compute_containers.append(ComputePipelineContainer(shader))
                elif shader.type == "render":
                    render_containers.append(RenderPipelineContainer(shader))
                else:
                    raise ValueError(f"Shader type {shader.type} is unknown.")
                if shader.needs_bake_function:
                    bake_functions.append(shader.bake_function)

            # Store results
            self.compute_containers = tuple(compute_containers)
            self.render_containers = tuple(render_containers)
            self.bake_functions = tuple(bake_functions)

        # If something has changed, update containers
        if changed:
            for container in self.compute_containers:
                container.update(wobject, renderstate, changed, tracker)
            for container in self.render_containers:
                container.update(wobject, renderstate, changed, tracker)


class PipelineContainer:
    """Object that wraps a set of wgpu pipeline objects for a single Shader object.

    A PipelineContainer is created for a specific combination of WorldObject and Renderstate
    (each differtent combination results in a different PipelineContainer).

    One shader results into multiple pipelines because of different render passes.

    The intermediate steps are also stored. When a dependency of a certain step
    changes (which we track) then only the steps below it need to be re-run.
    """

    def __init__(self, shader: ShaderInterface):
        self.shader = shader  # the corresponding ShaderInterface object
        self.shared = get_shared()  # the globally Shared object
        self.device = self.shared.device  # the global device

        # Dict to store info on the wobject that affects shaders or pipeline.
        # Fields are set in a tracking-context to make sure things update accordingly.
        self.wobject_info = {}

        # The info that the shader generates
        self.shader_hash = b""
        self.bindings_dicts = None  # dict of dict of bindings
        self.pipeline_info = None  # dict
        self.render_info = None  # dict

        # The wgpu objects that we generate.
        self.wgpu_shader = None
        self.wgpu_pipeline = None

        # The bindings map group_index -> group
        self.bind_group_layout_descriptors = []
        self.wgpu_bind_group_layouts = []
        self.wgpu_bind_groups = []

        # A flag to indicate that an error occurred and we cannot dispatch
        self.broken = False

    def _check_bindings(self):
        assert isinstance(self.bindings_dicts, dict), "bindings_dicts must be a dict"
        for key1, bindings_dict in self.bindings_dicts.items():
            assert isinstance(key1, int), f"bindings slot must be int, not {key1}"
            assert isinstance(bindings_dict, dict), "bindings_dict must be a dict"
            for key2, b in bindings_dict.items():
                assert isinstance(key2, int), f"bind group slot must be int, not {key2}"
                assert isinstance(b, Binding), f"binding must be Binding, not {b}"

    def update(self, wobject, renderstate, changed, tracker):
        """Make sure that the pipeline is up-to-date for the given renderstate."""

        # Ensure that the information provided by the shader is up-to-date
        if changed:
            try:
                self.update_shader_data(wobject, changed, tracker)
            except Exception as err:
                self.broken = 1
                raise err
            else:
                self.broken = 0

        # Ensure that the (renderstate specific) wgpu objects are up-to-date
        if not self.broken:
            try:
                self.update_wgpu_data(renderstate, changed)
            except Exception as err:
                self.broken = 2
                raise err
            else:
                self.broken = False

        if changed:
            logger.info(f"{wobject} shader update: {', '.join(sorted(changed))}.")

    def update_shader_data(self, wobject, changed, tracker):
        """Update the info that applies to all passes and renderstates."""

        if "create" in changed or "reset" in changed:
            self.wobject_info = {}
            with tracker.track_usage("reset"):
                self.wobject_info["pick_write"] = wobject.material.pick_write
                alpha_method = wobject.material.alpha_method
                self.wobject_info["alpha_method"] = alpha_method
                if alpha_method in ["opaque", "stochastic", "weighted"]:
                    self.wobject_info["alpha_config"] = wobject.material.alpha_config

            changed.update(("bindings", "pipeline_info", "render_info"))
            self.wgpu_shader = None

        if "bindings" in changed:
            with tracker.track_usage("!bindings"):
                self.bindings_dicts = self.shader.get_bindings_info(
                    wobject, self.shared
                )
            self._check_bindings()
            self.update_shader_hash()
            self.update_bind_groups()

        if "pipeline_info" in changed:
            with tracker.track_usage("pipeline_info"):
                self.pipeline_info = self.shader.get_pipeline_info(wobject, self.shared)
                self.wobject_info["depth_test"] = wobject.material.depth_test
                self.wobject_info["depth_compare"] = wobject.material.depth_compare
                self.wobject_info["depth_write"] = wobject.material.depth_write
                # For 'blended', the details need a new pipeline, but not a new shader
                self.wobject_info["alpha_config"] = wobject.material.alpha_config
            self._check_pipeline_info()
            changed.add("render_info")
            self.wgpu_pipeline = None

        if "render_info" in changed:
            with tracker.track_usage("render_info"):
                self.render_info = self.shader.get_render_info(wobject, self.shared)
            self._check_render_info()

    def update_wgpu_data(self, renderstate, changed):
        """Update the actual wgpu objects."""

        # Note: from here-on, we cannot access the wobject anymore, because any tracking should
        # be done in update_shader_data(). If you find that info on the wobject is needed,
        # add that info to self.wobject_info under the appropriate tracking context.

        # Update shaders
        if self.wgpu_shader is None:
            changed.add("shader")
            self.wgpu_shader = self._get_shader(renderstate)
            self.wgpu_pipeline = None

        # Update pipelines
        if self.wgpu_pipeline is None:
            changed.add("pipeline")
            self.wgpu_pipeline = self._compose_pipeline(renderstate)

    def update_shader_hash(self):
        """Update the shader hash, invalidating the wgpu shaders if it changed."""
        sh = self.shader.hash
        if sh != self.shader_hash:
            self.shader_hash = sh
            self.wgpu_shader = None

    def update_bind_groups(self):
        """
        - Calculate new bind_group_layout_descriptors (simple dicts).
        - When this has changed from our last version, we also update
          wgpu_bind_group_layouts and reset self.wgpu_pipeline
        - Calculate new wgpu_bind_groups.
        """

        # Check the bindings structure
        for i, bindings_dict in self.bindings_dicts.items():
            assert isinstance(i, int)
            assert isinstance(bindings_dict, dict)
            for j, b in bindings_dict.items():
                assert isinstance(j, int)
                assert isinstance(b, Binding)

        # Create two new dicts that correspond closely to the bindings.
        # Except this turns the dicts into lists.
        # These are the descriptors to create the wgpu bind groups and
        # bind group layouts.
        bg_descriptors = []
        bg_layout_descriptors = []
        for group_id in sorted(self.bindings_dicts.keys()):
            bindings_dict = self.bindings_dicts[group_id]
            while len(bg_descriptors) <= group_id:
                bg_descriptors.append([])
                bg_layout_descriptors.append([])
            bg_descriptor = bg_descriptors[group_id]
            bg_layout_descriptor = bg_layout_descriptors[group_id]
            for slot in sorted(bindings_dict.keys()):
                binding = bindings_dict[slot]
                binding_des, binding_layout_des = binding.get_bind_group_descriptors(
                    slot
                )
                bg_descriptor.append(binding_des)
                bg_layout_descriptor.append(binding_layout_des)

        # Clean up any trailing empty descriptors
        while bg_descriptors and not bg_descriptors[-1]:
            bg_descriptors.pop(-1)
            bg_layout_descriptors.pop(-1)

        # If the layout has changed, we need a new pipeline
        if self.bind_group_layout_descriptors != bg_layout_descriptors:
            self.bind_group_layout_descriptors = bg_layout_descriptors
            # Invalidate the pipeline
            self.wgpu_pipeline = None
            # Create wgpu objects for the bind group layouts
            self.wgpu_bind_group_layouts = []
            for bg_layout_descriptor in bg_layout_descriptors:
                bind_group_layout = get_cached_bind_group_layout(
                    self.device, bg_layout_descriptor
                )
                self.wgpu_bind_group_layouts.append(bind_group_layout)

        # Always create wgpu objects for the bind groups. Note that this dict
        # includes the buffer texture objects.
        self.wgpu_bind_groups = []
        for bg_descriptor, layout in zip(
            bg_descriptors, self.wgpu_bind_group_layouts, strict=True
        ):
            bind_group = get_cached_bind_group(self.device, layout, bg_descriptor)
            self.wgpu_bind_groups.append(bind_group)


class ComputePipelineContainer(PipelineContainer):
    """Container for compute pipelines."""

    # These checks are here to validate the output of the shader
    # methods, not for user code. So its ok to use assertions here.

    def _check_pipeline_info(self):
        pipeline_info = self.pipeline_info
        assert isinstance(pipeline_info, dict)

        assert set(pipeline_info.keys()) == set(), f"{pipeline_info.keys()}"

    def _check_render_info(self):
        render_info = self.render_info
        assert isinstance(render_info, dict)

        assert set(render_info.keys()) == {"indices"}, f"{render_info.keys()}"

        indices = render_info["indices"]
        assert isinstance(indices, (tuple, list))
        assert all(isinstance(i, int) for i in indices)
        assert len(indices) == 3

    def _get_shader(self, renderstate):
        """Get the wgpu shader module for the templated wgsl shade."""
        shader_kwargs = {}
        return get_cached_shader_module(self.device, self.shader, shader_kwargs)

    def _compose_pipeline(self, renderstate):
        """Create the wgpu pipeline object from the shader and bind group layouts."""

        # Create pipeline layout object from list of layouts
        pipeline_layout = get_cached_pipeline_layout(
            self.device, self.wgpu_bind_group_layouts
        )

        # Create pipeline object
        return get_cached_compute_pipeline(
            self.device, pipeline_layout, self.wgpu_shader
        )

    def dispatch(self, compute_pass):
        """Dispatch the pipeline, doing the actual compute job."""
        if self.broken:
            return

        # Collect what's needed
        pipeline = self.wgpu_pipeline
        indices = self.render_info["indices"]
        bind_groups = self.wgpu_bind_groups

        # Set pipeline and bindings
        compute_pass.set_pipeline(pipeline)
        for bind_group_id, bind_group in enumerate(bind_groups):
            compute_pass.set_bind_group(bind_group_id, bind_group, [], 0, 999999)

        # Compute!
        compute_pass.dispatch_workgroups(*indices)


class RenderPipelineContainer(PipelineContainer):
    """Container for render pipelines."""

    def __init__(self, shader):
        super().__init__(shader)
        self.strip_index_format = 0

    def _check_pipeline_info(self):
        pipeline_info = self.pipeline_info
        assert isinstance(pipeline_info, dict)

        expected = {"cull_mode", "primitive_topology"}
        assert set(pipeline_info.keys()) == expected, f"{pipeline_info.keys()}"
        self.update_strip_index_format()

    def _check_render_info(self):
        render_info = self.render_info
        assert isinstance(render_info, dict)

        expected = {"indices"}
        assert set(render_info.keys()) == expected, f"{render_info.keys()}"

        indices = render_info["indices"]
        assert isinstance(indices, (tuple, list))
        assert all(isinstance(i, int) for i in indices)
        assert len(indices) in (2, 4, 5)
        if len(indices) == 2:
            render_info["indices"] = indices[0], indices[1], 0, 0

    def update_strip_index_format(self):
        if not self.bindings_dicts or not self.pipeline_info:
            return
        # Set strip_index_format
        index_format = wgpu.IndexFormat.uint32
        strip_index_format = 0
        if "strip" in self.pipeline_info["primitive_topology"]:
            strip_index_format = index_format
        # Trigger a pipeline rebuild?
        if self.strip_index_format != strip_index_format:
            self.strip_index_format = strip_index_format
            self.wgpu_pipeline = None

    def _get_shader(self, renderstate):
        """Get the wgpu shader module for the templated shader."""
        blender = renderstate.blender
        renderstate_bind_group_index = len(self.wgpu_bind_groups)

        blender_kwargs = blender.get_shader_kwargs(
            self.wobject_info["pick_write"],
            self.wobject_info["alpha_config"],
        )
        renderstate_kwargs = renderstate.get_shader_kwargs(renderstate_bind_group_index)
        shader_kwargs = blender_kwargs.copy()
        shader_kwargs.update(renderstate_kwargs)

        return get_cached_shader_module(self.device, self.shader, shader_kwargs)

    def _compose_pipeline(self, renderstate):
        """Create the wgpu pipeline object from the shader, bind group layouts and other pipeline info."""

        strip_index_format = self.strip_index_format
        primitive_topology = self.pipeline_info["primitive_topology"]
        cull_mode = self.pipeline_info["cull_mode"]

        # Create pipeline layout object from list of layouts
        renderstate_bind_group_layout = renderstate.wgpu_bind_group_layout
        bind_group_layouts = [
            *self.wgpu_bind_group_layouts,
            renderstate_bind_group_layout,
        ]

        pipeline_layout = get_cached_pipeline_layout(self.device, bind_group_layouts)

        # Instantiate the pipeline objects.
        # Note: The pipeline relies on the color and depth descriptors, which
        # include the texture format and a few other static things.
        # This step should *not* rerun when e.g. the canvas resizes.
        blender = renderstate.blender
        depth_test = self.wobject_info["depth_test"]
        depth_compare = self.wobject_info["depth_compare"]
        depth_write = self.wobject_info["depth_write"]
        color_descriptors = blender.get_color_descriptors(
            self.wobject_info["pick_write"],
            self.wobject_info["alpha_config"],
        )
        depth_descriptor = blender.get_depth_descriptor(
            depth_test, depth_compare, depth_write
        )
        shader_module = self.wgpu_shader

        return get_cached_render_pipeline(
            self.device,
            pipeline_layout,
            shader_module,
            primitive_topology,
            strip_index_format,
            cull_mode,
            depth_descriptor,
            color_descriptors,
        )

    def draw(self, render_pass, renderstate):
        """Draw the pipeline, doing the actual rendering job."""
        if self.broken:
            return

        # Collect what's needed
        pipeline = self.wgpu_pipeline
        indices = self.render_info["indices"]
        bind_groups = self.wgpu_bind_groups

        # Set pipeline and bindings
        render_pass.set_pipeline(pipeline)

        for bind_group_id, bind_group in enumerate(bind_groups):
            render_pass.set_bind_group(bind_group_id, bind_group, [], 0, 99)

        renderstate_bind_group_id = len(bind_groups)
        renderstate_bind_group = renderstate.wgpu_bind_group
        render_pass.set_bind_group(
            renderstate_bind_group_id, renderstate_bind_group, [], 0, 99
        )

        # Draw!
        # draw(count_vertex, count_instance, first_vertex, first_instance)
        render_pass.draw(*indices)



"""
Containers for geometry data.

.. currentmodule:: pygfx.geometries

A geometry object contains the data that defines (the shape of) the object, such
as positions, plus data associated with these positions (normals, texcoords,
colors, etc.).

Its attributes are Buffer and Texture(View) objects. The data can be provided as
keyword arguments, which are converted to numpy arrays and wrapped in a Buffer
if necessary.

The names for these attributes are standardized so that the renderers know what
to expect. Each material requires certain attributes to be present, and may
support optional attributes. Optional attributes must always be "turned on" on
the material; their presence on the geometry does not mean that they're used.

The standardized names are:

* ``indices``: An index into per-vertex data. Typically Nx3 for mesh geometry.
* ``positions``: Nx3 positions (xyz), defining the location of e.g. vertices or
  points.
* ``normals``: Nx3 normal vectors. These may or may not be unit.
* ``texcoords``: Texture coordinates used to lookup the color for a vertex.
    Can be Nx1, Nx2 or Nx3, corresponding to a 1D, 2D and 3D texture map.
* ``colors``: Per-vertex colors. Must be NxM, with M 1-4 for gray,
    gray+alpha, rgb, rgba, respectively.
* ``sizes``: Scalar size per-vertex.
* ``grid``: A 2D or 3D Texture that contains a regular grid of data,
  i.e. for images and volumes.

.. rubric:: Basic Geometry
.. autosummary::
    :toctree: geometry/
    :template: ../_templates/custom_layout.rst

    box_geometry
    cylinder_geometry
    cone_geometry
    sphere_geometry
    plane_geometry
    Geometry

.. rubric:: Other/Special Geometry
.. autosummary::
    :toctree: geometry/
    :template: ../_templates/custom_layout.rst

    geometry_from_trimesh
    octahedron_geometry
    icosahedron_geometry
    dodecahedron_geometry
    tetrahedron_geometry
    torus_knot_geometry
    klein_bottle_geometry

"""

# ruff: noqa: F401

from ._base import Geometry
from ._box import box_geometry
from ._compat import geometry_from_trimesh, geometry_from_open3d
from ._cylinder import cylinder_geometry, cone_geometry
from ._sphere import sphere_geometry
from ._plane import plane_geometry, mobius_strip_geometry
from ._polyhedron import (
    octahedron_geometry,
    icosahedron_geometry,
    dodecahedron_geometry,
    tetrahedron_geometry,
)
from ._toroidal import torus_knot_geometry, klein_bottle_geometry

# Define __all__ for e.g. Sphinx
__all__ = [
    ob.__name__
    for ob in globals().values()
    if (isinstance(ob, type) and issubclass(ob, Geometry))
    or (callable(ob) and ob.__name__.endswith("_geometry"))
]
__all__.sort()
__all__.remove("Geometry")
__all__.insert(0, "Geometry")
__all__.append("geometry_from_trimesh")



from typing import Tuple

import numpy as np

from ._base import Controller


class PanZoomController(Controller):
    """A controller to pan and zoom a camera in a 2D plane  parallel to the screen.

    Default controls:

    * Left mouse button: pan.
    * Right mouse button: zoom (if `camera.maintain_aspect==False`, zooms in both dimensions).
    * Fourth mouse button: quickzoom
    * wheel: zoom to point.
    * alt+wheel: adjust fov.

    """

    _default_controls = {
        "mouse1": ("pan", "drag", (1, 1)),
        "mouse2": ("zoom", "drag", (0.005, -0.005)),
        "mouse4": ("quickzoom", "peek", 2),
        "wheel": ("zoom_to_point", "push", -0.001),
        "alt+wheel": ("fov", "push", -0.01),
    }

    def pan(self, delta: Tuple, rect: Tuple, *, animate=False):
        """Pan the camera (move relative to its local coordinate frame).

        If animate is True, the motion is damped. This requires the
        controller to receive events from the renderer/viewport.
        """

        # Note that we need that rect to determine mouse positions relative
        # to the viewport. If all events were adjusted to the viewport
        # than code like this would not have to care ...

        # Note how we return the result of _update_cameras when not animating.
        # If auto_update is False, _update_cameras will do nothing. The return value
        # is the camera state, which the calling code can then process further.

        if animate:
            action_tuple = ("pan", "push", (1.0, 1.0))
            action = self._create_action(None, action_tuple, (0.0, 0.0), None, rect)
            action.set_target(delta)
            action.done = True
        elif self._cameras:
            vecx, vecy = self._get_camera_vecs(rect)
            self._update_pan(delta, vecx=vecx, vecy=vecy)
            return self._update_cameras()

    def _update_pan(self, delta, *, vecx, vecy):
        # These update methods all accept one positional arg: the delta.
        # it can additionally require keyword args, from a set of names
        # that new actions store. These include:
        # rect, screen_pos, vecx, vecy

        assert isinstance(delta, tuple) and len(delta) == 2

        cam_state = self._get_camera_state()
        position = cam_state["position"]

        # Update position, panning left means dragging the scene to the
        # left, i.e. move the camera to the right, thus the minus. But
        # since screen pixels go from top to bottom, while the camera's
        # up vector points ... up, the y component is negated twice.
        new_position = position - vecx * delta[0] + vecy * delta[1]
        self._set_camera_state({"position": new_position})

    def zoom(self, delta: Tuple, rect: Tuple, *, animate=False):
        """Zoom the view with the given amount.

        The delta can be either a scalar or 2-element tuple. The zoom
        multiplier is calculated using ``2**delta``. If the camera has
        maintain_aspect set to True, only the second value is used.

        Note that the camera's distance, width, and height are adjusted,
        not its zoom property.

        If animate is True, the motion is damped. This requires the
        controller to receive events from the renderer/viewport.
        """

        if animate:
            action_tuple = ("zoom", "push", (1.0, 1.0))
            action = self._create_action(None, action_tuple, (0.0, 0.0), None, rect)
            action.set_target(delta)
            action.done = True
        elif self._cameras:
            self._update_zoom(delta)
            return self._update_cameras()

    def _update_zoom(self, delta):
        if isinstance(delta, (int, float)):
            delta = (delta, delta)
        assert isinstance(delta, tuple) and len(delta) == 2

        fx = 2 ** delta[0]
        fy = 2 ** delta[1]
        new_cam_state = self._zoom(fx, fy, self._get_camera_state())
        self._set_camera_state(new_cam_state)

    def zoom_to_point(self, delta: float, pos: Tuple, rect: Tuple, *, animate=False):
        """Zoom the view while panning to keep the position under the cursor fixed.

        If animate is True, the motion is damped. This requires the
        controller to receive events from the renderer/viewport.
        """

        if animate:
            action_tuple = ("zoom_to_point", "push", 1.0)
            action = self._create_action(None, action_tuple, (0.0, 0.0), None, rect)
            action.set_target(delta)
            action.done = True
        elif self._cameras:
            self._update_zoom_to_point(delta, screen_pos=pos, rect=rect)
            return self._update_cameras()

    def _update_zoom_to_point(self, delta, *, screen_pos, rect):
        if isinstance(delta, tuple) and len(delta) == 2:
            delta = delta[1]
        assert isinstance(delta, (int, float))

        # Actually only zoom in one direction
        fy = 2**delta

        new_cam_state = self._zoom(fy, fy, self._get_camera_state())
        self._set_camera_state(new_cam_state)

        pan_delta = self._get_panning_to_compensate_zoom(fy, screen_pos, rect)
        vecx, vecy = self._get_camera_vecs(rect)
        self._update_pan(pan_delta, vecx=vecx, vecy=vecy)

    def _zoom(self, fx, fy, cam_state):
        position = cam_state["position"]
        maintain_aspect = cam_state["maintain_aspect"]
        width = cam_state["width"]
        height = cam_state["height"]
        extent = 0.5 * (width + height)

        # Scale width and height equally, or use width and height.
        if maintain_aspect:
            new_width = width / fy
            new_height = height / fy
        else:
            new_width = width / fx
            new_height = height / fy

        # Get new position
        new_extent = 0.5 * (new_width + new_height)
        pos2target1 = self._get_target_vec(cam_state, extent=extent)
        pos2target2 = self._get_target_vec(cam_state, extent=new_extent)
        new_position = position + pos2target1 - pos2target2

        return {
            "width": new_width,
            "height": new_height,
            "position": new_position,
            "fov": cam_state["fov"],
        }

    def _get_panning_to_compensate_zoom(self, multiplier, screen_pos, rect):
        # Get viewport info
        x, y, w, h = rect

        # Distance from the center of the rect
        delta_screen_x = screen_pos[0] - x - w / 2
        delta_screen_y = screen_pos[1] - y - h / 2
        delta_screen1 = np.array([delta_screen_x, delta_screen_y])

        # New position after zooming
        delta_screen2 = delta_screen1 * multiplier

        # The amount to pan is the difference, but also scaled with the multiplier
        # because pixels take more/less space now.
        return tuple((delta_screen1 - delta_screen2) / multiplier)



import numpy as np

from ._base import Geometry


def sphere_geometry(
    radius=1,
    width_segments=32,
    height_segments=16,
    phi_start=0,
    phi_length=np.pi * 2,
    theta_start=0,
    theta_length=np.pi,
):
    """Generate a sphere.

    Creates a sphere that has its center in the local origin. The sphere is
    constructed by creating a grid of longituidnal and latitudinal lines and
    placing vetices at the intersections. The area between 4 vertices is then
    approximated by a rectangular-shaped face.

    Optionally, the geometry can be limited to a segment of the full sphere
    described by a longituidnal and latitudinal arc.

    Parameters
    ----------
    radius : float
        The radius of the sphere. Vertices are placed at this distance around
        the local origin.
    width_segments : int
        The number of (evenly-spaced) longitudinal lines to draw within the
        selected sphere segment.
    height_segments : int
        The number of (evenly-spaced) latitudinal lines to draw within the
        selected sphere segment.
    phi_start : float
        The angle (in rad) at which to start the longitudinal lines. Zero means
        the first line is drawn in the xz-plane.
    phi_length : float
        The central angle (in rad) of the sphere segments latitudinal segment.
    theta_start : float
        The angle (in rad) at which to start the latitudinal lines. Zero means
        the first line is drawn at the "tip" of the sphere, i.e., at ``(0, 0,
        radius)``.
    theta_length : float
        The central angle (in rad) of the sphere segment's longitudinal segment.

    Returns
    -------
    sphere : Geometry
        A geometry object that represents the requested sphere or sphere segment.
        Mathematically, it is an open orientable manifold.

    """

    # create grid of spherical coordinates
    nx = width_segments + 1
    phi_end = phi_start + phi_length
    phi = np.linspace(phi_start, phi_end, num=nx, dtype=np.float32)

    ny = height_segments + 1
    theta_end = theta_start + theta_length
    theta = np.linspace(theta_start, theta_end, num=ny, dtype=np.float32)

    # grid has shape (ny, nx)
    phi_grid, theta_grid = np.meshgrid(phi, theta)

    # convert to cartesian coordinates
    theta_grid_sin = np.sin(theta_grid)
    xx = np.cos(phi_grid) * theta_grid_sin * -1
    yy = np.cos(theta_grid)
    zz = np.sin(phi_grid) * theta_grid_sin

    # construct normals and positions
    normals = np.stack([xx, yy, zz], axis=-1)
    positions = normals * radius

    # construct texture coordinates
    # u maps 0..1 to phi_start..phi_start+phi_length
    # v maps 1..0 to theta_start..theta_start+theta_length
    uu = (phi_grid - phi_start) / phi_length
    vv = 1 - ((theta_grid - theta_start) / theta_length)
    texcoords = np.stack([uu, vv], axis=-1)

    # the face indices
    # assign an index to every vertex on the grid
    idx = np.arange(nx * ny, dtype=np.uint32).reshape((ny, nx))
    # for every panel (height_segments, width_segments) there is a quad (2, 3)
    indices = np.empty((height_segments, width_segments, 2, 3), dtype=np.uint32)
    # create a grid of initial idx for the panels
    indices[:, :, 0, 0] = idx[
        np.arange(height_segments)[:, None], np.arange(width_segments)[None, :]
    ]
    # the remainder of the indices for every panel are relative
    indices[:, :, 0, 1] = indices[:, :, 0, 0] + nx
    indices[:, :, 0, 2] = indices[:, :, 0, 0] + 1
    indices[:, :, 1, 0] = indices[:, :, 0, 0] + nx + 1
    indices[:, :, 1, 1] = indices[:, :, 1, 0] - nx
    indices[:, :, 1, 2] = indices[:, :, 1, 0] - 1

    return Geometry(
        indices=indices.reshape((-1, 3)),
        positions=positions.reshape((-1, 3)),
        normals=normals.reshape((-1, 3)),
        texcoords=texcoords.reshape((-1, 2)),
    )



import numpy as np


class Bounds:
    """Represent bounds using an axis-aligned bounding box (aabb) plus a radius.

    The radius represents a sphere centered at the aabb. Semantically,
    the bounds are the intersection of the box and the spere. This
    allows applying transforms without growing the bounds because of
    the box's corners.

    When the radius is None, or near-equal to half the diagonal of the aabb,
    the bounds respresents a box. The radius is at least the
    max(width,height,depth) of the aabb, i.e. only cutting off its corners.
    """

    __slots__ = ["aabb", "radius"]

    def __init__(self, aabb, radius=None):
        aabb = np.asarray(aabb)
        if aabb.shape != (2, 3):
            raise ValueError("aabb must be 2x3 array")
        self.aabb = aabb
        if radius is not None:
            min_radius = 0.5 * float(abs((aabb[1] - aabb[0]).max()))  # w/h/d
            max_radius = 0.5 * float(np.linalg.norm(aabb[1] - aabb[0]))  # diagonal
            if radius > 0.99 * max_radius:
                radius = None
            else:
                radius = max(radius, min_radius)
        self.radius = radius

    def __repr__(self):
        p1 = ", ".join(f"{i:0.4g}" for i in self.aabb[0])
        p2 = ", ".join(f"{i:0.4g}" for i in self.aabb[1])
        r = "without radius"
        if self.radius is not None:
            r = f"with radius {self.radius:0.4g}"
        return f"<Bounds aabb from ({p1}) to ({p2}) {r} at {hex(id(self))}>"

    def __eq__(self, other):
        return np.all(self.aabb == other.aabb) and self.radius == other.radius

    @property
    def center(self):
        aabb = self.aabb
        return tuple(0.5 * (aabb[0] + aabb[1]))

    @property
    def sphere(self):
        aabb = self.aabb
        radius = self.radius
        if radius is None:
            radius = 0.5 * float(np.linalg.norm(aabb[1] - aabb[0]))
        c = 0.5 * (aabb[0] + aabb[1])
        return float(c[0]), float(c[1]), float(c[2]), radius

    @classmethod
    def from_points(cls, points, calculate_radius=False):
        aabb, radius = points_to_bounds(points, calculate_radius)
        if aabb is None:
            return None
        return cls(aabb, radius)


def points_to_bounds(points, calculate_radius=True):
    # Check points
    points = np.asarray(points)
    if not (points.ndim == 2 and points.shape[1] in (2, 3)):
        raise ValueError("Points must be a list of 2D or 3D points.")
    if points.shape[0] == 0:
        return None, None

    # Get aabb
    aabb = np.array([np.min(points, axis=0), np.max(points, axis=0)], dtype=float)
    if aabb.shape[1] == 2:
        aabb = np.column_stack([aabb, np.zeros((2, 1), float)])

    # If the aabb has nonfinite values, fix and recurse! The min/max above
    # is a very efficient way to detect whether there are nonfinite values.
    # So we don't have to first do a full check on the points themselves.
    if not np.isfinite(aabb).all():
        finite_mask = np.isfinite(points).all(axis=1)
        return points_to_bounds(points[finite_mask])

    radius = None
    if calculate_radius:
        # Find radius
        c = 0.5 * (aabb[0] + aabb[1])
        distances = np.linalg.norm(points - c, axis=1)
        radius = float(max(distances))
        # The radius only has effect if its smaller than half the diagonal
        diagonal = float(np.linalg.norm(aabb[1] - aabb[0]))
        if radius > 0.49 * diagonal:
            radius = None

    return aabb, radius



"""
The blender is the part of the renderer that manages output targets and alpha_config.
"""

import wgpu  # only for flags/enums

from .effectpasses import create_full_quad_pipeline
from .shared import get_shared
from ..wgsl import load_wgsl


standard_texture_des = {
    "sample_type": wgpu.TextureSampleType.unfilterable_float,
    "view_dimension": wgpu.TextureViewDimension.d2,
    "multisampled": False,
}


def blend_dict(src_factor, dst_factor, operation):
    return {
        "operation": operation,
        "src_factor": src_factor,
        "dst_factor": dst_factor,
    }


DEPTH_MAP = {
    "": wgpu.CompareFunction.always,
    "<": wgpu.CompareFunction.less,
    "<=": wgpu.CompareFunction.less_equal,
    "==": wgpu.CompareFunction.equal,
    "!=": wgpu.CompareFunction.not_equal,
    ">=": wgpu.CompareFunction.greater_equal,
    ">": wgpu.CompareFunction.greater,
}

usg = wgpu.TextureUsage
default_targets = {
    # Inside all our shaders, as well as in the EffectPass shaders, we follow a
    # "linear workflow"; inside the shaders all colors are assumed to be in linear
    # space (colors read from colormaps are converted when read). The final color,
    # displayed on screen (or stored to PNG, etc.) should be sRGB so that the colors
    # are perceived correctly. This conversion should preferably be done at the end.
    #
    # As for storing intermediate results, using 8bit colors in linear space would
    # mean that we lose precision in regions where the human eye is sensitive.
    # Ideally we'd use 'rgba16float', and maybe we should, it's 2025, but let's
    # change that in a self-contained commit, because it will affect memory usage,
    # and needs a feature?
    #
    # Using 'rgba8unorm_srgb' means that in each post-effect step, the colors are
    # stored in srgb, and auto-conveted to linear when read/written. This seems a
    # bit weird, because from the pov of math that works on linear values, it loses
    # precision in a non-linear way. But that way we lose precision "in a way that's
    # linear to the human eye", which results in a better end-result than using
    # 'rgba8unorm'. Also see https://github.com/pmndrs/postprocessing
    #
    # This is 4 bytes per pixel.
    # TODO: use half-floats
    "color": (
        wgpu.TextureFormat.rgba8unorm_srgb,
        usg.RENDER_ATTACHMENT | usg.COPY_SRC | usg.TEXTURE_BINDING,
    ),
    # A texture of the same size, to allow post-processing effects.
    # When applying effects, the color and altcolor texture are ping-ponged.
    "altcolor": (
        wgpu.TextureFormat.rgba8unorm_srgb,
        usg.RENDER_ATTACHMENT | usg.COPY_SRC | usg.TEXTURE_BINDING,
    ),
    # The depth buffer should preferably at least 24bit - we need that precision. It's 4 bytes per pixel.
    # 32 but is cool, but we may want stencil at some point, so depth24plus_stencil8 seems like a good default.
    # The depth24plus is either depth32float or depth24unorm, depending on the backend.
    # Note that there is also depth32float-stencil8, but it needs the (webgpu) extension with the same name.
    "depth": (
        wgpu.TextureFormat.depth32float,
        usg.RENDER_ATTACHMENT | usg.COPY_SRC | usg.TEXTURE_BINDING,
    ),
    # The pick texture has 4 16bit channels. It's 8 bytes per pixel.
    # These bits are divided over the pick data, e.g. 20 for the wobject id.
    "pick": (
        wgpu.TextureFormat.rgba16uint,
        usg.RENDER_ATTACHMENT | usg.COPY_SRC,
    ),
    # The accumulation buffer collects weighted fragments. It's 8 bytes per pixel.
    "accum": (
        wgpu.TextureFormat.rgba16float,
        usg.RENDER_ATTACHMENT | usg.TEXTURE_BINDING,
    ),
    # The reveal buffer collects the weights. It's 1 byte per pixel.
    # McGuire: "Using R16F for the revealage render target will give slightly better
    # precision and make it easier to tune the algorithm, but a 2x savings on bandwidth
    # and memory footprint for that texture may make it worth compressing into R8 format."
    "reveal": (
        wgpu.TextureFormat.r8unorm,
        usg.RENDER_ATTACHMENT | usg.TEXTURE_BINDING,
    ),
}


class Blender:
    """Manage how fragments are blended and end up in the final target.
    Each renderer has one blender object.
    """

    def __init__(self, *, enable_pick=True, enable_depth=True):
        self.device = get_shared().device

        # We could allow custom targets, but this is not yet implemented in the methods.
        # The code in this init shows the first steps of what that could look like.
        custom_targets = {}  # name -> (format, usage)  could allow users to specify this

        # The size (2D in pixels) of the textures.
        self.size = (0, 0)

        # Objects for the combination pass
        self._weighted_blending_was_used_in_last_pass = False
        self._weighted_resolve_pass_pipeline = None
        self._weighted_resolve_pass_bind_group = None

        # A dict that contains the metadata for all render targets.
        self._texture_info = {}  # name -> dict

        # A dict with the actual textures
        self._textures = {}  # name -> Texture

        # -- setup targets

        # Collect render targets
        all_targets = default_targets.copy()
        for name in sorted(custom_targets):
            all_targets[name] = custom_targets[name]
        if not enable_pick:
            all_targets.pop("pick")
        if not enable_depth:
            all_targets.pop("depth")

        # Convert to final dict
        for name, (format, usage) in all_targets.items():
            assert isinstance(name, str)
            assert format in wgpu.TextureFormat
            assert isinstance(usage, int)
            self._texture_info[name] = {
                "name": name,
                "format": format,
                "usage": usage,
                "is_used": False,
                "clear": True,
            }

    @property
    def hash(self):
        """The hash for this blender.

        This is used by the renderstate. If an object is rendered with two renderers,
        we still want this to work, even if the corresponding blenders results in incompatible pipelines.

        Use cases are a blender without a pick or depth target. Or blenders with different color formats.

        In such cases, their blender hashes will be different, resulting in a separate pipeline object for each.
        Similar to rendering with a different number of lights will result in separate pipeline objects.
        """
        return ", ".join(
            info["name"] + ": " + info["format"] for info in self._texture_info.values()
        )

    @property
    def texture_info(self):
        """A dict of dicts, containing per-rendertarget info."""
        return self._texture_info

    def get_texture(self, name):
        """Get the texture object for the given name."""
        return self._textures.get(name)

    def get_texture_view(self, name, usage, *, create_if_not_exist=False):
        """Get a texture view for the given name, with the given usage. Returns None if the texture does not exist."""
        texture = self._textures.get(name)

        if texture is None:
            if not create_if_not_exist:
                return None
            else:
                texinfo = self._texture_info[name]
                texture = self.device.create_texture(
                    dimension="2d",
                    size=(*self.size, 1),
                    usage=texinfo["usage"],
                    format=texinfo["format"],
                )
            self._textures[name] = texture

        return texture.create_view(usage=usage)

    def _get_texture_view_for_rendering(self, name):
        """Get a texture view for the given name with rendert-attachment usage. Creates the texture if it does not exist."""
        return self.get_texture_view(
            name, wgpu.TextureUsage.RENDER_ATTACHMENT, create_if_not_exist=True
        )

    def clear(self):
        """Clear all the buffers (on the next time they're attached)."""
        for texstate in self._texture_info.values():
            texstate["clear"] = True

    def ensure_target_size(self, size):
        """Make sure that the textures are resized if necessary."""

        assert len(size) == 2
        size = int(size[0]), int(size[1])
        if size == self.size:
            return

        # Set new size
        self.size = size

        # Any bind group is now invalid because they include source textures.
        self._weighted_resolve_pass_bind_group = None

        # All textures are invalidated too
        self._textures = {}

    def get_color_descriptors(self, material_pick_write, alpha_config):
        """Get the dict color-descriptors that pipeline.py needs to create the render pipeline.

        Called per object when the pipeline is created.
        """
        bf, bo = wgpu.BlendFactor, wgpu.BlendOperation

        alpha_method = alpha_config["method"]

        # Get the color_blend mini-dict
        if alpha_method == "opaque":
            color_blend = {
                "alpha": blend_dict(bf.one, bf.zero, bo.add),
                "color": blend_dict(bf.one, bf.zero, bo.add),
            }
        elif alpha_method == "blended":
            # Note that color_constant and alpha_constant are not yet supported.
            # We'd need to call GPURenderPassEncoder.set_blend_constant(rgba), close to where we call GPURenderPassEncoder.set_viewport() in renderer.py
            color_blend = {
                "color": blend_dict(
                    alpha_config["color_src"],
                    alpha_config["color_dst"],
                    alpha_config.get("color_op", "add"),
                ),
                "alpha": blend_dict(
                    alpha_config["alpha_src"],
                    alpha_config["alpha_dst"],
                    alpha_config.get("alpha_op", "add"),
                ),
            }
        elif alpha_method == "stochastic":
            color_blend = {
                "alpha": blend_dict(bf.one, bf.zero, bo.add),
                "color": blend_dict(bf.one, bf.zero, bo.add),
            }
        elif alpha_method == "weighted":
            color_blend = None  # handled below
        else:
            raise RuntimeError(f"Unexpected alpha_method {alpha_method!r}")

        # Build target state dicts
        color_target_state = {
            "name": "color",
            "blend": color_blend,
            "write_mask": wgpu.ColorWrite.ALL,
        }
        target_states = [color_target_state]

        # More work for 'weighted' method5
        if alpha_method == "weighted":
            accum_target_state = {
                "name": "accum",
                "blend": {
                    "alpha": blend_dict(bf.one, bf.one, bo.add),
                    "color": blend_dict(bf.one, bf.one, bo.add),
                },
                "write_mask": wgpu.ColorWrite.ALL,
            }
            reveal_target_state = {
                "name": "reveal",
                "blend": {
                    "alpha": blend_dict(bf.zero, bf.one_minus_src_alpha, bo.add),
                    "color": blend_dict(bf.zero, bf.one_minus_src, bo.add),
                },
                "write_mask": wgpu.ColorWrite.ALL,
            }
            target_states = [accum_target_state, reveal_target_state]

        # Add pick target if the blender supports it. All pipelines must have matching (target states in their) pipelines.
        # Whether or not the pick target is used is determined by the material using the write_mask.
        if "pick" in self._texture_info:
            target_state = {
                "name": "pick",
                "blend": None,
                "write_mask": wgpu.ColorWrite.ALL if material_pick_write else 0,
            }
            target_states.append(target_state)

        # Note: in a similar way we could allow custom additional render targets.

        # Set format for each target, and register what targets were used
        for target_state in target_states:
            name = target_state.pop("name")
            texinfo = self._texture_info[name]
            target_state["format"] = texinfo["format"]
            texinfo["is_used"] = True

        return target_states

    def get_depth_descriptor(self, depth_test, depth_compare, depth_write):
        """Get the dict depth-descriptors that pipeline.py needs to create the render pipeline.

        Called per object when the pipeline is created.
        """

        if "depth" not in self._texture_info:
            return None

        depth_write = bool(depth_write)
        if depth_test:
            wgpu_depth_compare = DEPTH_MAP[depth_compare]
        else:
            wgpu_depth_compare = wgpu.CompareFunction.always

        texinfo = self._texture_info["depth"]
        texinfo["is_used"] = True

        return {
            "format": texinfo["format"],
            "depth_write_enabled": depth_write,
            "depth_compare": wgpu_depth_compare,
            "stencil_read_mask": 0,
            "stencil_write_mask": 0,
            "stencil_front": {},  # use defaults
            "stencil_back": {},  # use defaults
        }

    def get_color_attachments(self, pass_type):
        """Get the texture render targets for color.

        These are dynamically attached right before rendering a batch of objects.
        This is called by the renderer for each 'batch' of objects (of a particular pass_type) is being rendered.
        """

        color_attachment = {
            "name": "color",
            "resolve_target": None,
            "clear_value": (0, 0, 0, 0),
            "load_op": wgpu.LoadOp.load,  # maybe set to clear at the end of this func
            "store_op": wgpu.StoreOp.store,
        }
        pick_attachment = {
            "name": "pick",
            "resolve_target": None,
            "clear_value": (0, 0, 0, 0),
            "load_op": wgpu.LoadOp.load,
            "store_op": wgpu.StoreOp.store,
        }
        attachments = [color_attachment]

        self._weighted_blending_was_used_in_last_pass = False

        if pass_type == "weighted":
            self._weighted_blending_was_used_in_last_pass = True
            # We always clear the textures at the beginning of a pass, because at the end of
            # that pass it will be merged with the color buffer using the combine pass.
            accum_clear_value = 0, 0, 0, 0
            reveal_clear_value = 1, 0, 0, 0
            accum_attachment = {
                "name": "accum",
                "resolve_target": None,
                "clear_value": accum_clear_value,
                "load_op": wgpu.LoadOp.load,
                "store_op": wgpu.StoreOp.store,
            }
            reveal_attachment = {
                "name": "reveal",
                "resolve_target": None,
                "clear_value": reveal_clear_value,
                "load_op": wgpu.LoadOp.load,
                "store_op": wgpu.StoreOp.store,
            }
            attachments = [accum_attachment, reveal_attachment]

        # Add pick attachment if this blender supports picking.
        if "pick" in self._texture_info:
            attachments.append(pick_attachment)

        # Finalize attachments
        for attachment in attachments:
            name = attachment.pop("name")
            texinfo = self._texture_info[name]

            if texinfo["clear"]:
                texinfo["clear"] = False
                attachment["load_op"] = wgpu.LoadOp.clear

            attachment["view"] = self._get_texture_view_for_rendering(name)

        return attachments

    def get_depth_attachment(self):
        """Get the texture render targets for depth/stencil. These are dynamically attached right before rendering a batch of objects."""

        if "depth" not in self._texture_info:
            return None

        texinfo = self._texture_info["depth"]

        load_op = wgpu.LoadOp.load
        if texinfo["clear"]:
            texinfo["clear"] = False
            load_op = wgpu.LoadOp.clear

        view = self._get_texture_view_for_rendering("depth")

        # We don't use the stencil yet, but when we do, we will also have to specify
        # "stencil_read_only", "stencil_load_op", and "stencil_store_op"
        return {
            "view": view,
            "depth_clear_value": 1.0,
            "depth_load_op": load_op,
            "depth_store_op": wgpu.StoreOp.store,
        }

    def get_shader_kwargs(self, material_pick_write, alpha_config):
        """Get the shader templating variables for the given alpha config."""

        alpha_method = alpha_config["method"]

        if alpha_method == "opaque":
            make_rgb = "color.rgb"
            if alpha_config["premultiply_alpha"]:
                make_rgb = "color.rgb * color.a"

            fragment_output_code = """
            struct FragmentOutput {
                // virtualfield stub: bool = false;
                @location(0) color: vec4<f32>,
                MAYBE_PICK@location(1) pick: vec4<u32>,
            };

            fn apply_virtual_fields_of_fragment_output(outp: ptr<function,FragmentOutput>, stub: bool) {
                let color = (*outp).color;
                let rgb = MAKE_RGB;
                (*outp).color = vec4f(rgb, 1.0);  // force alpha 1
            }
            """.replace("MAKE_RGB", make_rgb)

        elif alpha_method == "blended":
            fragment_output_code = """
            struct FragmentOutput {
                @location(0) color: vec4<f32>,
                MAYBE_PICK@location(1) pick: vec4<u32>,
            };
            """

        elif alpha_method == "stochastic":
            # We want the seed for the random function to be such that the
            # result is deterministic, so that rendered images can be visually
            # compared. We also (in general) want the seed to be stable for e.g.
            # movement of the mouse or object.
            #
            # So, let's *not* use:
            #
            # * The global-id: because it's determined by how many objects have
            #   been created earlier (in the current process).
            # * Object position, because then it will flicker when it moves.
            # * varyings.position.z, because then it will flicker when the
            #   camera moves.
            #
            # We can use these:
            #
            # * varyings.position.xy (the screen coord) is a common seed. It
            #   will give an effect often observed in games, where the pattern
            #   is "stuck to the screen".
            # * the renderer-id to distinguish objects while being reproducable
            #   in a scene.
            # * varyings.elementIndex to prevent two regions of the same object
            #   to get the same pattern, e.g. two points in a points object.

            seed = alpha_config.get("seed", "screen").lower()
            upos = "upos" + str({"screen": 0, "object": 1, "element": 2}[seed])
            pattern = alpha_config.get("pattern", "blue").lower().replace("_", "-")
            if pattern in ("blue-noise", "blue"):
                random_call = f"blueNoise2({upos})"
            elif pattern in ("white-noise", "white"):
                random_call = f"whiteNoise({upos})"
            elif pattern in ("bayer8", "bayer"):
                # Using upos2 becomes too noisy, but upos1 (i.e. including the per-object seed) seems quite alright!
                random_call = f"bayerPattern({upos})"
            else:
                raise ValueError(f"Unexpected stochastic pattern: {pattern!r}")

            fragment_output_code = load_wgsl("noise.wgsl")  # cannot use include here

            fragment_output_code += """
            struct FragmentOutput {
                // virtualfield position: vec3f = varyings.position.xyz;
                // virtualfield objectId: u32 = u_wobject.renderer_id;
                // virtualfield elementIndex: u32 = varyings.elementIndex;
                @location(0) color: vec4<f32>,
                MAYBE_PICK@location(1) pick: vec4<u 32>,
            };

            fn apply_virtual_fields_of_fragment_output(outp: ptr<function,FragmentOutput>, position: vec3f, objectId: u32, elementIndex: u32) {

                // Early exit
                let alpha = (*outp).color.a;
                if (alpha >= 1.0) { return; }

                let screenSize = u_stdinfo.physical_size.xy;

                // Compose seeds
                let seed1 = hashu(objectId);
                let seed2 = hashu(objectId) ^ hashu(elementIndex+1);

                // Compose positions
                let upos0 = vec2u(position.xy);
                let upos1 = upos0 + vec2u(seed1 >> 16, seed1 & 0xffff);
                let upos2 = upos0 + vec2u(seed2 >> 16, seed2 & 0xffff);

                // Generate a random number with a blue-noise distribution, i.e. resulting in uniformly sampled points with very little 'structure'
                // Blue noise has is great for sampling problems like this, because it has few low-frequency components, so the noise is very 'fine'.
                // The bayer pattern looks nicer than the noise, but is not suited for mixing multiple transparent layers.
                var rand = 0.0;

                if true {  // set to false to use split-screen debug mode
                    rand = RANDOM_CALL;
                } else if position.x < 0.5 * screenSize.x {
                    //rand = blueNoise2(upos2);
                    //rand = random(position.x * position.y * position.z);  // more or less original white noise version
                    rand = bayerPattern(upos1);
                } else {
                    rand = blueNoise2(upos2);
                }

                // Render or drop the fragment?
                if ( alpha < 1.0 - ALPHA_COMPARE_EPSILON && alpha < rand ) { discard; }
                (*outp).color.a = 1.0;  // fragments that pass through are opaque
            }
            """.replace("RANDOM_CALL", random_call)

        elif alpha_method == "weighted":
            use_alpha = "alpha", "use_alpha", "weighted_blending_use_alpha"
            weight_default = alpha_config.get("weight", "alpha")
            if weight_default in use_alpha:
                weight_default = "weighted_blending_use_alpha"
            alpha_default = alpha_config.get("alpha", "alpha")
            if alpha_default in use_alpha:
                alpha_default = "weighted_blending_use_alpha"

            fragment_output_code = """
            struct FragmentOutput {
                // virtualfield color : vec4<f32> = vec4<f32>(0.0)
                // virtualfield alpha : f32 = ALPHA_DEFAULT
                // virtualfield weight : f32 = WEIGHT_DEFAULT
                @location(0) accum: vec4<f32>,
                @location(1) reveal: f32,
                MAYBE_PICK@location(2) pick: vec4<u32>,
            };
            const weighted_blending_use_alpha: f32 = -42.0;

            fn apply_virtual_fields_of_fragment_output(outp: ptr<function,FragmentOutput>, color: vec4<f32>,  _alpha: f32, _weight: f32) {
                let alpha = select(_alpha, color.a, _alpha == weighted_blending_use_alpha);
                let weight = select(_weight, color.a, _weight == weighted_blending_use_alpha);
                (*outp).accum = vec4<f32>(color.rgb * alpha, alpha) * weight;
                (*outp).reveal = alpha;  // yes, alpha, not weight
            }
            """.replace("WEIGHT_DEFAULT", weight_default).replace(
                "ALPHA_DEFAULT", alpha_default
            )

            # Note 1: could also take user-specified transmittance into account.
            # Note 2: its also possible to undo a fragment contribution. For this the accum
            # and reveal buffer must be float to avoid clamping. And we'd do `abs(color.a)` above.
            # The formula would then be:
            #    out.accum = - out.accum;
            #    out.reveal = 1.0 - 1.0 / (1.0 - alpha);

        else:
            raise RuntimeError(f"Unexpected alpha_method {alpha_method!r}")

        # Enable/disable picking in the shader
        do_pick = material_pick_write and "pick" in self._texture_info
        fragment_output_code = fragment_output_code.replace(
            "MAYBE_PICK", "" if do_pick else "// "
        )

        return {
            "alpha_method": alpha_method,
            "fragment_output_code": fragment_output_code,
            "write_pick": do_pick,
        }

    def perform_weighted_resolve_pass(self, command_encoder):
        """Perform a render-pass to resolve the result of weighted blending."""

        # This is only needed if objects were rendered with weighted blending
        if not self._weighted_blending_was_used_in_last_pass:
            return

        # Get bindgroup and pipeline. The creation should only happens once per blender lifetime.
        if not self._weighted_resolve_pass_pipeline:
            self._weighted_resolve_pass_pipeline = self._create_combination_pipeline()
        if not self._weighted_resolve_pass_pipeline:
            return []

        # Get the bind group. A new one is needed when the source textures resize.
        if not self._weighted_resolve_pass_bind_group:
            self._weighted_resolve_pass_bind_group = (
                self._create_combination_bind_group(
                    self._weighted_resolve_pass_pipeline.get_bind_group_layout(0)
                )
            )

        # Get info on the color target texture
        texinfo = self._texture_info["color"]
        load_op = wgpu.LoadOp.load
        if texinfo["clear"]:
            texinfo["clear"] = False
            load_op = wgpu.LoadOp.clear

        # Render
        render_pass = command_encoder.begin_render_pass(
            color_attachments=[
                {
                    "view": self._get_texture_view_for_rendering("color"),
                    "resolve_target": None,
                    "load_op": load_op,
                    "store_op": wgpu.StoreOp.store,
                }
            ],
            depth_stencil_attachment=None,
            occlusion_query_set=None,
        )
        render_pass.set_pipeline(self._weighted_resolve_pass_pipeline)
        render_pass.set_bind_group(0, self._weighted_resolve_pass_bind_group, [], 0, 99)
        render_pass.draw(4, 1)
        render_pass.end()

    def _create_combination_pipeline(self):
        binding_layout = [
            {
                "binding": 0,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "texture": standard_texture_des,
            },
            {
                "binding": 1,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "texture": standard_texture_des,
            },
        ]

        bf, bo = wgpu.BlendFactor, wgpu.BlendOperation
        targets = [
            {
                "format": self._texture_info["color"]["format"],
                "blend": {
                    "alpha": blend_dict(bf.one, bf.one_minus_src_alpha, bo.add),
                    "color": blend_dict(bf.one, bf.one_minus_src_alpha, bo.add),
                },
            },
        ]

        wgsl = """
            @group(0) @binding(0)
            var r_accum: texture_2d<f32>;
            @group(0) @binding(1)
            var r_reveal: texture_2d<f32>;

            @fragment
            fn fs_main(varyings: Varyings) -> @location(0) vec4<f32> {
                let epsilon = 1e-6;
                let texIndex = vec2i(varyings.position.xy);

                // Sample
                let accum = textureLoad(r_accum, texIndex, 0).rgba;
                let reveal = textureLoad(r_reveal, texIndex, 0).r;

                // Exit if no transparent fragments was written
                // This discard does not brake early-z, because with weighted blending you're rendering each object anyway.
                if (reveal >= 1.0) { discard; }  // no transparent fragments here

                // Reconstruct the color and alpha, and set final color, with premultiplied alpha
                let avgColor = accum.rgb / max(accum.a, epsilon);
                let alpha = 1.0 - reveal;
                return vec4<f32>(avgColor * alpha, alpha);
            }
        """

        return create_full_quad_pipeline(targets, binding_layout, wgsl)

    def _create_combination_bind_group(self, bind_group_layout):
        # This must match the binding_layout above
        bind_group_entries = [
            {
                "binding": 0,
                "resource": self.get_texture_view(
                    "accum", wgpu.TextureUsage.TEXTURE_BINDING
                ),
            },
            {
                "binding": 1,
                "resource": self.get_texture_view(
                    "reveal", wgpu.TextureUsage.TEXTURE_BINDING
                ),
            },
        ]
        return self.device.create_bind_group(
            layout=bind_group_layout, entries=bind_group_entries
        )



import os
import json

from .. import logger, get_resources_dir
from ._fontfinder import FontFile, get_all_fonts, weight_dict, style_dict


# Allow "slanted" to be requested. It will request a regular font,
# and the glyphs will be slanted in the shader.
style_dict = style_dict.copy()
style_dict["slanted"] = "slanted"


class FontProps:
    """
    An object for storing font properties. Typically used as a request for an actual font.

    Parameters:
        family (str, tuple): The name of a font, or a list of font names
            in decreasing order of priority.
        style (str): Either 'normal', 'italic' or 'oblique'.
        weight (float, str): A numeric value in the range 100-900 or one of
            'ultralight', 'light', 'normal', 'regular', 'medium',
            'semibold', 'demibold', 'bold', 'extra bold', 'black'.
    """

    def __init__(
        self,
        family=None,
        *,
        style=None,
        weight=None,
    ):
        # Check family
        if family is None:
            family = font_manager.default_font_props.family
        elif isinstance(family, str):
            family = (family,)
        elif isinstance(family, (tuple, list)):
            for x in family:
                if not isinstance(x, str):
                    cls = type(x).__name__
                    raise TypeError(f"Font family must be str, not '{cls}'")
            family = tuple(family)
        else:
            cls = type(family).__name__
            raise TypeError(f"Font family must be str or tuple-of-str, not '{cls}'")

        # Check style
        if style is None:
            style = font_manager.default_font_props.style
        elif isinstance(style, str):
            try:
                style = style_dict[style.lower()]
            except KeyError:
                raise TypeError(f"Style string not known: '{style}'") from None
        else:
            cls = type(style).__name__
            raise TypeError("Font style must be str, not '{cls}'")

        # Check weight
        if weight is None:
            weight = font_manager.default_font_props.weight
        elif isinstance(weight, str):
            try:
                weight = weight_dict[weight.lower()]
            except KeyError:
                raise TypeError(f"Weight string not known: '{weight}'") from None
        elif isinstance(weight, int):
            weight = min(900, max(100, weight))
        else:
            raise TypeError("Weight must be an int (100-900) or a string.")

        self._kwargs = {
            "family": family,
            "style": style,
            "weight": weight,
        }

    def __repr__(self):
        fam = ", ".join(repr(x) for x in self.family)
        return f"<FontProps {fam} at 0x{hex(id(self))}>"

    def copy(self, **kwargs):
        """Make a copy of the font prop, with given kwargs replaced."""
        d = self._kwargs.copy()
        for k, v in kwargs.items():
            if v is not None:
                d[k] = v
        return self.__class__(**d)

    @property
    def family(self):
        """A tuple of font family strings, e.g. "Noto Sans" or "Arial"."""
        return self._kwargs["family"]

    @property
    def style(self):
        """The style, one of "normal", "italic", or "oblique"."""
        return self._kwargs["style"]

    @property
    def weight(self):
        """The font weight, as a number between 100-900."""
        return self._kwargs["weight"]


class FontManager:
    """Storage and discovery of text glyphs.

    The main purpose of the font manager is font selection, i.e. selecting a
    font based on the text to be rendered and a font_props object.

    This manager uses a default font set based on the Noto fonts, allowing a
    very complete Unicode coverage, including rare and ancient scripts. Users
    probably don't have the full set of Noto fonts installed. When a font is
    missing that the manager knows would support the text, a useful error
    message is produced, that includes a link to where the font can be installed
    from.

    There is a singleton instance of this class at ``pygfx.utils.text.font_manager``.
    """

    def __init__(self):
        self._default_font_props = FontProps((), style="normal", weight="regular")
        self._warned_for_codepoints = set()
        self._family_to_font = {}  # name -> style -> FontFile
        self._load_default_font_index()
        self._load_fonts()

    @property
    def default_font_props(self):
        """The default font properties."""
        # Note: could implement set_default_font(font_props) at some point ...
        return self._default_font_props

    def add_font_file(self, font_file):
        """Add the given font_file to the collection of fonts. The
        font_file can be a filename or a FontFile object. Returns the
        FontFile object for the font.
        """
        # Obtain FontFile object
        if isinstance(font_file, FontFile):
            ff = font_file
        elif isinstance(font_file, str):
            ff = FontFile(font_file)
        else:
            raise TypeError("add_font_file() expects FontFile or str filename.")

        # Select on family name
        variants = self._family_to_font.setdefault(ff.family, {})

        # Warn for duplicates
        if ff.variant in variants:
            old = variants[ff.variant]
            logger.debug(f"Duplicate font {ff.name} ({old.filename} -> {ff.filename})")

        # Store
        variants[ff.variant] = ff
        return ff

    def _load_fonts(self):
        # Populate the dict of font objects
        for ff in get_all_fonts():
            self.add_font_file(ff)

        # The main font of the default font is the fallback of fallbacks.
        # We copy the fontfile so we can detect when it's used to show tofu's.
        ff = self._family_to_font["Noto Sans"]["Regular"]
        self._fallback_font = FontFile(ff.filename, ff.family, ff.variant)

    def _load_default_font_index(self):
        # Load the json
        index_filename = os.path.join(get_resources_dir(), "noto_default_index.json")
        with open(index_filename, "rt", encoding="utf-8") as f:
            ob = json.load(f)
        families, fnames, index = ob["families"], ob["filenames"], ob["index"]

        # Create a map to lookup the fname from the family
        self._default_family_to_fname = {
            family: fname for family, fname in zip(families, fnames, strict=True)
        }

        # Create a dict that maps codepoint to tuple of names
        self._default_font_map = {}
        for k, v in index.items():
            codepoint = int(k)
            self._default_font_map[codepoint] = tuple(families[i] for i in v)

    def get_fonts(self):
        """Get a list of all registered FontFile objects. E.g. to show a list
        of all available fonts:
        ``for ff in font_manager.get_fonts(): print(ff.family, "-", ff.variant)``
        """
        fonts = []
        for family in sorted(self._family_to_font.keys()):
            for ff in self._family_to_font[family].values():
                fonts.append(ff)
        return fonts

    def select_font(self, text, font_props):
        """Select the (best) fonts for the given text. Returns a list of (text, font_file)
        tuples, because different characters in the text may require different fonts.
        """

        # The selection strategy depends on whether preferred fonts are
        # given. If this is the case, the preferred font should be used
        # where possible. This applies per character. For the remaining
        # text, the default font is used. In this case, any font from
        # the default set is fine. We just try to make the pieces as
        # long as possible. For characters that we cannot render, the
        # fallback font is used - we need *a* font to render the tofu's ...

        fallback_font = self._fallback_font

        # Select fonts that match the preferred font_props
        preferred_fonts = []
        for family in font_props.family:
            # Get variants for the given family
            try:
                variants = self._family_to_font[family]
            except KeyError:
                continue
            # Select best variant
            best_score, best_ff = 0, None
            for ff in variants.values():
                score = (
                    int(font_props.style == ff.style)
                    + 600
                    - abs(ff.weight - font_props.weight)
                )
                if score > best_score:
                    best_score, best_ff = score, ff
            # Select
            if best_ff:
                preferred_fonts.append(best_ff)

        # First apply the most-preferred font to each character.
        # If not preferred_fonts are supplied, we can take a shortcut.
        text_pieces1 = []
        if preferred_fonts:
            codepoint = ord(text[0])
            last_font = self._select_preferred_font_for_codepoint(
                codepoint, preferred_fonts
            )
            last_i = i = 0
            for i in range(1, len(text)):
                codepoint = ord(text[i])
                font = self._select_preferred_font_for_codepoint(
                    codepoint, preferred_fonts
                )
                if font is not last_font:
                    text_pieces1.append((text[last_i:i], last_font))
                    last_i = i
                    last_font = font
            text_pieces1.append((text[last_i : i + 1], last_font))
        else:
            text_pieces1.append((text, None))

        failed_codepoints = []

        # Now process the pieces that don't have a font yet, using the default fonts.
        text_pieces2 = []
        for text, font in text_pieces1:
            if font is not None:
                text_pieces2.append((text, font))
            else:
                codepoint = ord(text[0])
                fonts = self._select_default_fonts_for_codepoint(codepoint)
                last_i = i = 0
                for i in range(1, len(text)):
                    codepoint = ord(text[i])
                    # Quick continue for default font
                    if len(fonts) == 1 and fonts[0].has_codepoint(codepoint):
                        continue
                    new_fonts = [ff for ff in fonts if ff.has_codepoint(codepoint)]
                    if new_fonts:
                        # Our selection of fonts is still nonzero
                        fonts = new_fonts
                    else:
                        if not fonts:
                            failed_codepoints.append(codepoint)
                            fonts = self._select_default_fonts_for_codepoint(codepoint)
                            if not fonts:
                                continue
                            last_font = fallback_font
                        else:
                            last_font = fonts[0]
                            fonts = self._select_default_fonts_for_codepoint(codepoint)
                        text_pieces2.append((text[last_i:i], last_font))
                        last_i = i
                last_font = fonts[0] if fonts else fallback_font
                if last_i == 0:
                    text_pieces2.append((text, last_font))
                else:
                    text_pieces2.append((text[last_i : i + 1], last_font))

        # Did we encounter characters that we cannot render?
        failed_texts = [text for text, font in text_pieces2 if font is fallback_font]
        if failed_texts:
            codepoints = {ord(c) for c in "".join(failed_texts)}
            if any(cp not in self._warned_for_codepoints for cp in codepoints):
                self._warned_for_codepoints.update(codepoints)
                logger.warning(self._produce_font_warning(*failed_texts))

        return text_pieces2

    def _select_preferred_font_for_codepoint(self, codepoint, preferred_fonts):
        # Select one font, in order of preference
        for ff in preferred_fonts:
            if ff.has_codepoint(codepoint):
                return ff
        return None

    def _select_default_fonts_for_codepoint(self, codepoint):
        # Select all fonts capable of rendering this character
        fonts = []
        default_families = self._default_font_map.get(codepoint, ())
        for family in default_families:
            try:
                ff = self._family_to_font[family]["Regular"]
            except KeyError:
                continue
            if ff.has_codepoint(codepoint):
                fonts.append(ff)
        return fonts

    def _produce_font_warning(self, *failed_texts):
        # Get the codepoints that failed
        codepoints = list({ord(c) for c in "".join(failed_texts)})

        # Collect families. We try to find a set of families that supports all
        # the failed characters. In theory we could analyse things more and
        # show the minimal set of fonts that supports all chars.
        all_families = set()
        min_families = set(self._default_font_map.get(codepoints[0], ()))
        for cp in codepoints:
            families = self._default_font_map.get(cp, ())
            all_families.update(families)
            min_families.intersection_update(families)

        # Define what to report
        msg = f"Cannot render chars '{' '.join(failed_texts)}'. "
        fonts_to_link = set()
        if min_families:
            fonts_to_link = min_families
            if len(min_families) == 1:
                msg += "To fix this, install the following font:"
            else:
                msg += "To fix this, install any of the following fonts:"
        elif all_families:
            fonts_to_link = all_families
            msg += "To fix this, install (some) of the following fonts:"
        else:
            msg += "Even the Noto font set does not support these characters."

        # Show links
        msg += "\n"
        for family in sorted(fonts_to_link):
            fname = self._default_family_to_fname[family]
            msg += f"    https://pygfx.github.io/noto-mirror/#{fname}\n"

        # We return a string, making this method easy to test
        return msg


# Instantiate the global/default font manager
font_manager = FontManager()



"""
The main renderer class. This class wraps a canvas or texture and it
manages the rendering process.
"""

import os
import time
import logging
from typing import Literal

import numpy as np
import wgpu
import pylinalg as la
from rendercanvas import BaseRenderCanvas

from ....objects._base import id_provider
from ....objects import (
    KeyboardEvent,
    RootEventHandler,
    PointerEvent,
    WheelEvent,
    WindowEvent,
    WorldObject,
)
from ....objects._lights import (
    Light,
    PointLight,
    DirectionalLight,
    SpotLight,
    AmbientLight,
)
from ....cameras import Camera
from ....resources import Texture
from ....resources._base import resource_update_registry
from ....utils import Color
from ....utils.enums import PixelFilter

from ... import Renderer
from .blender import Blender
from .effectpasses import EffectPass, OutputPass, PPAAPass, FXAAPass, DDAAPass
from .pipeline import get_pipeline_container_group
from .update import update_resource, ensure_wgpu_object
from .shared import get_shared
from .renderstate import get_renderstate
from .shadowutil import render_shadow_maps
from .mipmapsutil import generate_texture_mipmaps
from .utils import GfxTextureView


logger = logging.getLogger("pygfx")


class WobjectWrapper:
    """To temporary wrap each wobject for each draw."""

    __slots__ = ["pass_type", "render_containers", "sort_key", "wobject"]

    def __init__(self, wobject, sort_key, pass_type):
        self.wobject = wobject
        self.sort_key = sort_key
        self.pass_type = pass_type
        self.render_containers = None


class FlatScene:
    def __init__(self, scene, view_matrix=None, object_count=0):
        self._view_matrix = view_matrix
        self._wobject_wrappers = []  # WobjectWrapper's
        self.lights = {
            "point_lights": [],
            "directional_lights": [],
            "spot_lights": [],
            "ambient_color": [0, 0, 0],
        }
        self.shadow_objects = []
        self.object_count = object_count
        self.add_scene(scene)

    def _iter_scene(self, ob, render_order=0):
        if not ob.visible:
            return
        render_order += ob.render_order
        yield ob, render_order
        for child in ob._children:
            yield from self._iter_scene(child, render_order)

    def add_scene(self, scene):
        """Add a scene to the total flat scene. Is usually called just once."""

        # Put some attributes as vars in this namespace for faster access
        view_matrix = self._view_matrix
        wobject_wrappers = self._wobject_wrappers

        for wobject, render_order in self._iter_scene(scene):
            # Dereference the object in case its a weak proxy
            wobject = wobject._self()
            # Assign renderer id's
            self.object_count += wobject._assign_renderer_id(self.object_count + 1)
            # Update things like transform and uniform buffers
            wobject._update_object()

            # Light objects
            if isinstance(wobject, Light):
                if isinstance(wobject, PointLight):
                    self.lights["point_lights"].append(wobject)
                elif isinstance(wobject, DirectionalLight):
                    self.lights["directional_lights"].append(wobject)
                elif isinstance(wobject, SpotLight):
                    self.lights["spot_lights"].append(wobject)
                elif isinstance(wobject, AmbientLight):
                    r, g, b = wobject.color.to_physical()
                    ambient_color = self.lights["ambient_color"]
                    ambient_color[0] += r * wobject.intensity
                    ambient_color[1] += g * wobject.intensity
                    ambient_color[2] += b * wobject.intensity

            # Shadowable objects
            if wobject.cast_shadow and wobject.geometry is not None:
                self.shadow_objects.append(wobject)

            # Renderable objects
            material = wobject._material
            if material is not None:
                render_queue = material.render_queue
                alpha_method = material.alpha_method

                # By default sort back-to-front, for correct blending.
                dist_sort_sign = -1
                # But for opaque queues, render front-to-back to avoid overdraw.
                if 1500 < render_queue <= 2500:
                    dist_sort_sign = 1

                pass_type = alpha_method

                # Get depth sorting flag. Note that use camera's view matrix, since the projection does not affect the depth order.
                # It also means we can set projection=False optimalization.
                # Also note that we look down -z.
                if alpha_method == "weighted":
                    dist_flag = 0
                elif view_matrix is None:
                    dist_flag = -1
                else:
                    relative_pos = la.vec_transform(
                        wobject.world.position, view_matrix, projection=False
                    )
                    # Cam looks towards -z: negate to get distance
                    distance_to_camera = float(-relative_pos[2])
                    dist_flag = distance_to_camera * dist_sort_sign

                sort_key = (render_queue, render_order, dist_flag)
                wobject_wrappers.append(WobjectWrapper(wobject, sort_key, pass_type))

    def sort(self):
        """Sort the world objects."""
        self._wobject_wrappers.sort(key=lambda ob: ob.sort_key)

    def collect_pipelines_container_groups(self, renderstate):
        """Select and resolve the pipeline, compiling shaders, building pipelines and composing binding as needed."""
        self._compute_pipeline_containers = compute_pipeline_containers = []
        self._bake_functions = bake_functions = []
        for wrapper in self._wobject_wrappers:
            container_group = get_pipeline_container_group(wrapper.wobject, renderstate)
            compute_pipeline_containers.extend(container_group.compute_containers)
            wrapper.render_containers = container_group.render_containers
            for func in container_group.bake_functions:
                bake_functions.append((wrapper.wobject, func))

    def call_bake_functions(self, camera, logical_size):
        """Call any collected bake functions."""
        # Enable pipelines to update data on the CPU. This usually includes
        # baking data into buffers. This is CPU intensive, but in practice
        # it is only used by a few materials.
        for wobject, func in self._bake_functions:
            func(wobject, camera, logical_size)

    def iter_compute_pipelines(self):
        """Generator that yields the collected compute pipelines."""
        for pipeline_container in self._compute_pipeline_containers:
            yield pipeline_container

    def iter_render_pipelines_per_pass_type(self):
        """Generator that yields (pass_type, wobjects), with pass_type 'opaque', 'transparency' or 'weighted'."""
        current_pass_type = ""
        current_pipeline_containers = []
        for wrapper in self._wobject_wrappers:
            if wrapper.pass_type != current_pass_type:
                if current_pipeline_containers:
                    yield (current_pass_type, current_pipeline_containers)
                current_pass_type = wrapper.pass_type
                current_pipeline_containers = []
            current_pipeline_containers.extend(wrapper.render_containers)
        if current_pipeline_containers:
            yield (current_pass_type, current_pipeline_containers)


class WgpuRenderer(RootEventHandler, Renderer):
    """Turns Scenes into rasterized images using wgpu.

    Parameters
    ----------
    target : WgpuCanvas or Texture
        The target to render to. It is also used to determine the size of the
        render buffer.
    pixel_scale : float, optional
        The scale between the internal resolution and the physical resolution of the canvas.
        Setting to None (default) selects 1 if the screens looks to be HiDPI and 2 otherwise.
    pixel_ratio : float, optional
        The ratio between the number of internal pixels versus the logical pixels on the canvas.
        If both ``pixel_ratio`` and ``pixel_scale`` are set, ``pixel_ratio`` is ignored.
    pixel_filter : str, PixelFilter, optional
        The type of interpolation / reconstruction filter to use. Default 'mitchell'.
    show_fps : bool
        Whether to display the frames per second. Beware that
        depending on the GUI toolkit, the canvas may impose a frame rate limit.
    sort_objects : bool
        If True, sort objects by depth before rendering. If False, the
        rendering order is mainly based on the objects ``render_order`` and position
        in the scene graph.
    enable_events : bool
        If True, forward wgpu events to pygfx's event system.
    gamma_correction : float
        The gamma correction to apply in the final render stage. Typically a
        number between 0.0 and 2.0. A value of 1.0 indicates no correction.
    ppaa : str, optional
        The post-processing anti-aliasing to apply: "default", "none", "fxaa", "ddaa".
        By default it resolves to "ddaa".
    """

    def __init__(
        self,
        target,
        *args,
        pixel_scale=None,
        pixel_ratio=None,
        pixel_filter: PixelFilter = "mitchell",
        show_fps=False,
        sort_objects=True,
        enable_events=True,
        gamma_correction=1.0,
        ppaa="default",
        **kwargs,
    ):
        blend_mode = kwargs.pop("blend_mode", None)
        super().__init__(*args, **kwargs)

        # blend_mode is deprecated; raise error with somewhat helpful message
        if blend_mode:
            self.blend_mode = blend_mode

        # Check and normalize inputs
        if not isinstance(target, (Texture, GfxTextureView, BaseRenderCanvas)):
            raise TypeError(
                f"Render target must be a Canvas or Texture, not a {target.__class__.__name__}"
            )
        self._target = target
        self.pixel_ratio = pixel_ratio
        if pixel_scale is not None:
            self.pixel_scale = pixel_scale

        # Make sure we have a shared object (the first renderer creates the instance)
        self._shared = get_shared()
        self._device = self._shared.device

        # Count number of objects encountered
        self._object_count = 0

        # Init counter to auto-clear
        self._renders_since_last_flush = 0

        # Cache renderstate objects for n draws
        self._renderstates_per_flush = []

        # Get target format
        self.gamma_correction = gamma_correction
        self._gamma_correction_srgb = 1.0
        if isinstance(target, BaseRenderCanvas):
            self._canvas_context = self._target.get_context("wgpu")
            # Select output format. We currently don't have a way of knowing
            # what formats are available, so if not srgb, we gamma-correct in shader.
            target_format = self._canvas_context.get_preferred_format(
                self._shared.adapter
            )
            if not target_format.endswith("srgb"):
                self._gamma_correction_srgb = 1 / 2.2  # poor man's srgb
            # Also configure the canvas
            self._canvas_context.configure(
                device=self._device,
                format=target_format,
            )
        else:
            # Also enable the texture for render and display usage
            self._target._wgpu_usage |= wgpu.TextureUsage.RENDER_ATTACHMENT
            self._target._wgpu_usage |= wgpu.TextureUsage.TEXTURE_BINDING

        self._blender = Blender()
        self._effect_passes = ()
        self.ppaa = ppaa
        self._name_of_texture_with_effects = (
            None  # none, or the blender's name of the texture
        )

        self.sort_objects = sort_objects

        # Prepare object that performs the final render step into a texture
        self._output_pass = OutputPass()
        self.pixel_filter = pixel_filter

        # Initialize a small buffer to read pixel info into
        # Make it 256 bytes just in case (for bytes_per_row)
        self._pixel_info_buffer = self._device.create_buffer(
            size=16,
            usage=wgpu.BufferUsage.COPY_DST | wgpu.BufferUsage.MAP_READ,
        )

        # Init fps measurements
        self._show_fps = bool(show_fps)
        now = time.perf_counter()
        self._fps = {"start": now, "count": 0}

        if enable_events:
            self.enable_events()

    @property
    def device(self):
        """A reference to the global wgpu device."""
        return self._device

    @property
    def target(self):
        """The render target. Can be a canvas, texture or texture view."""
        return self._target

    @property
    def pixel_scale(self) -> float:
        """The scale between the internal resolution and the physical resolution of the canvas.

        * If the scale is 1, the internal texture has the same size as the target.
        * If the scale is larger than 1, you're doing SSAA.
        * If the scale is smaller than 1, you're rendering at a low resolution, and then upscaling the result.

        Note that a ``pixel_scale`` of 1 or 2 is more performant than fractional values.

        Setting this value to ``None``, will select a hirez configuration: It
        selects 1 if the target looks like a HiDPI screen (i.e.
        ``canvas.pixel_ratio>=2``), and 2 otherwise. That way, the internal
        texture size is the same, regardless of the user's system/monitor.
        """
        return self._pixel_scale

    @pixel_scale.setter
    def pixel_scale(self, pixel_scale: None | int | float):
        if pixel_scale is None:
            # Select hirez config
            self._pixel_scale = 2.0  # default
            if isinstance(self._target, BaseRenderCanvas):
                target_pixel_ratio = self._target.get_pixel_ratio()
                if target_pixel_ratio >= 2.0:
                    self._pixel_scale = 1.0
        else:
            pixel_scale = float(pixel_scale)
            if pixel_scale < 0.1 or pixel_scale > 10:
                raise ValueError("renderer.pixel_scale must be bwteen 0.1 and 10.")
            self._pixel_scale = pixel_scale

    @property
    def pixel_ratio(self) -> float:
        """The ratio between the number of internal pixels versus the logical pixels on the canvas.

        ``pixel_ratio = pixel_scale * canvas.pixel_ratio``

        Setting this prop also changes the ``pixel_scale``. This can be used to
        configure the size of the internal texture relative to the canvas'
        *logical* size.

        Setting this value to ``None`` is the same as setting ``pixel_scale`` to None,
        and results in a ``pixel_ratio`` of at least 2.

        Note that setting ``pixel_ratio`` to 2.0 does not have the same effect, because the
        canvas pixel_ratio can be e.g. 1.5, in which case the resulting ``pixel_scale`` becomes fractional.
        """
        target_pixel_ratio = 1
        if isinstance(self._target, BaseRenderCanvas):
            target_pixel_ratio = self._target.get_pixel_ratio()
        return self._pixel_scale * target_pixel_ratio

    @pixel_ratio.setter
    def pixel_ratio(self, pixel_ratio: None | float):
        if pixel_ratio is None:
            self.pixel_scale = None
        else:
            target_pixel_ratio = 1
            if isinstance(self._target, BaseRenderCanvas):
                target_pixel_ratio = self._target.get_pixel_ratio()
            pixel_scale = pixel_ratio / target_pixel_ratio
            if 0.9 < pixel_scale < 1.1:
                pixel_scale = 1  # snap
            elif 1.9 < pixel_scale < 2.1:
                pixel_scale = 2  # snap
            self.pixel_scale = pixel_scale

    @property
    def pixel_filter(self) -> PixelFilter:
        """The type of interpolation / reconstruction filter to use when flushing the result.

        See :obj:`pygfx.utils.enums.PixelFilter`.

        The renderer renders everything to an internal texture, which,
        depending on the ``pixel_scale``, may have a different physical size than
        the target (i.e. canvas). In the process of rendering the result
        to the target, a filter is applied, resulting in SSAA if the
        target size is smaller, and upsampling when the target size is larger.
        When the internal texture has the same size as the target, no filter is applied (equivalent to nearest).

        The filter defines how the interpolation is done (when the source and target are not of the same size).
        """
        return self._output_pass.filter

    @pixel_filter.setter
    def pixel_filter(self, value: PixelFilter):
        # For backwards compatibility, allow 0, 1, 0.0, 1.0, False, and True.
        if value == 0:
            value = "nearest"
        elif value == 1:
            value = "mitchell"
        if not isinstance(value, str):
            raise TypeError("Pixel filter must be a str.")
        value = value.lower()
        if value not in PixelFilter.__args__:
            raise ValueError(
                f"Pixel filter must be one of {PixelFilter}, not {value!r}"
            )
        self._output_pass.filter = value
        self._pixel_filter = value

    @property
    def ppaa(
        self,
    ) -> Literal["default", "none", "fxaa", "ddaa"]:
        """The post-processing anti-aliasing to apply.

        * "default": use the value specified by ``PYGFX_DEFAULT_PPAA``, defaulting to "ddaa".
        * "none": do not apply aliasing.
        * "fxaa": applies Fast Approxomate AA, a common method.
        * "ddaa": applies Directional Diffusion AA, a modern improved method.

        The ``PYGFX_DEFAULT_PPAA`` environment variable can e.g. be set to "none" for image tests,
        so that the image tests don't fail when we update the ddaa method.

        Note that SSAA can be achieved by using a pixel_scale > 1. This can be well combined with PPAA,
        since the PPAA is applied before downsampling to the target texture.
        """
        ppaa = "none"
        for effect_pass in self._effect_passes:
            if isinstance(effect_pass, PPAAPass):
                ppaa = effect_pass.__class__.__name__.split("Pass")[0].lower()
        return ppaa

    @ppaa.setter
    def ppaa(self, ppaa: Literal["default", "none", "fxaa", "ddaa"]):
        if not isinstance(ppaa, str):
            raise TypeError(f"renderer.ppaa must be a string, not {ppaa!r}")

        # Collect list of effect passes that are not a ppaa
        effect_passes = []
        for effect_pass in self._effect_passes:
            if not isinstance(effect_pass, PPAAPass):
                effect_passes.append(effect_pass)

        # Handle default
        algorithm = ppaa.lower()
        if algorithm == "default":
            algorithm = os.getenv("PYGFX_DEFAULT_PPAA", "").lower()
            if not algorithm or algorithm == "default":
                algorithm = "ddaa"

        if algorithm == "none":
            pass  # don't add a pass
        elif algorithm == "ddaa":
            effect_passes.append(DDAAPass())
        elif algorithm == "fxaa":
            effect_passes.append(FXAAPass())
        else:
            raise ValueError(f"Invalid value for renderer.ppaa: {ppaa!r}")

        self.effect_passes = effect_passes

    @property
    def rect(self):
        """The rectangular viewport for the renderer area."""
        return (0, 0, *self.logical_size)

    @property
    def logical_size(self):
        """The size of the render target in logical pixels."""
        target = self._target
        if isinstance(target, BaseRenderCanvas):
            return target.get_logical_size()
        elif isinstance(target, Texture):
            return target.size[:2]  # assuming pixel-ratio 1
        else:
            raise TypeError(f"Unexpected render target {target.__class__.__name__}")

    @property
    def physical_size(self):
        """The physical size of the internal render texture."""
        target = self._target
        if isinstance(self._target, BaseRenderCanvas):
            target_physical_size = self._target.get_physical_size()
        else:
            target_physical_size = target.size[:2]
        w, h = target_physical_size
        pixel_scale = self._pixel_scale
        return max(1, int(w * pixel_scale)), max(1, int(h * pixel_scale))

    @property
    def blend_mode(self):
        raise DeprecationWarning(
            "renderer.blend_mode is removed. Use material.alpha_mode instead."
        )

    @blend_mode.setter
    def blend_mode(self, value):
        raise DeprecationWarning(
            "renderer.blend_mode is removed. Use material.alpha_mode instead."
        )

    @property
    def sort_objects(self):
        """Whether to sort world objects by depth before rendering. Default False.

        By default, the render order is defined by:

          1. the object's ``render_order`` property;
          2. whether the object is opaque/transparent/weighted/unknown;
          3. the object's distance to the camera;
          4. the object's position in the scene graph (based on a depth-first search).

        If ``sort_objects`` is ``False``, step 3 (sorting using the camera transform) is omitted.
        """
        return self._sort_objects

    @sort_objects.setter
    def sort_objects(self, value):
        self._sort_objects = bool(value)

    @property
    def gamma_correction(self):
        """The gamma correction applied in the final composition step."""
        return self._gamma_correction

    @gamma_correction.setter
    def gamma_correction(self, value):
        self._gamma_correction = 1.0 if value is None else float(value)
        if isinstance(self._target, BaseRenderCanvas):
            self._target.request_draw()

    @property
    def effect_passes(self):
        """A tuple of ``EffectPass`` instances.

        Together they form a chain of post-processing passes to produce the final visual result.
        They are executed in order when the rendered image is flushed to the target (e.g. the screen).
        """
        return self._effect_passes

    @effect_passes.setter
    def effect_passes(self, effect_passes):
        effect_passes = tuple(effect_passes)
        for step in effect_passes:
            if not isinstance(step, EffectPass):
                raise TypeError(
                    f"A renderer effect-pass step must be an instance of EffectPass, not {step!r}"
                )
        self._effect_passes = effect_passes

    def clear(self, *, all=False, color=False, depth=False, weights=False):
        """Clear one or more of the render targets.

        Users typically don't need to use this method. But sometimes it can be convenient to e.g.
        render a scene, and then clear the depth before rendering another scene.

        * all: clear all render targets; a fully clean sheeth.
        * color: clear the color buffer to rgba all zeros.
        * depth: clear the depth buffer.
        * weights: clear the render targets for weighted blending (the accum and reveal textures).

        """
        if not (all or color or depth or weights):
            raise ValueError(
                "renderer.clear() needs at least all, color, depth, or weights set to True."
            )

        if all:
            self._blender.clear()
        else:
            if color:
                self._blender.texture_info["color"]["clear"] = True
            if depth:
                self._blender.texture_info["depth"]["clear"] = True
            if weights:
                self._blender.texture_info["accum"]["clear"] = True
                self._blender.texture_info["reveal"]["clear"] = True

    def render(
        self,
        scene: WorldObject,
        camera: Camera,
        *,
        rect=None,
        clear=None,
        flush=True,
        # deprecated:
        clear_color=None,
    ):
        """Render a scene with the specified camera as the viewpoint.

        Parameters:
            scene (WorldObject): The scene to render, a WorldObject that
                optionally has child objects.
            camera (Camera): The camera object to use, which defines the
                viewpoint and view transform.
            rect (tuple, optional): The rectangular region to draw into,
                expressed in logical pixels, a.k.a. the viewport.
            clear (bool, optional): Whether to clear the color and depth buffers
                before rendering. By default this is True on the first
                call to ``render()`` after a flush, and False otherwise.
            flush (bool, optional): Whether to flush the rendered result into
                the target (texture or canvas). Default True.
        """

        # A good time to reset this flag.
        self._name_of_texture_with_effects = None

        # Manage stored renderstate objects. Each renderstate object used will be stored at least a few draws.
        if self._renders_since_last_flush == 0:
            self._renderstates_per_flush.insert(0, [])
            self._renderstates_per_flush[16:] = []

        # Define whether to clear render targets
        if clear_color is not None:
            logger.warning(
                "renderer.render(.. clear_color) is deprecated in favor of 'clear'."
            )
        if clear is None:
            clear = self._renders_since_last_flush == 0
        if clear:
            self._blender.clear()

        self._renders_since_last_flush += 1

        # todo: also note that the fragment shader is (should be) optional
        #      (e.g. depth only passes like shadow mapping or z prepass)

        # Get size for the render textures
        logical_size = self.logical_size
        physical_size = self.physical_size
        if not all(i > 0 for i in logical_size):
            return
        pixel_ratio = physical_size[1] / logical_size[1]

        # Update the render targets
        self._blender.ensure_target_size(physical_size)

        # Get viewport in physical pixels
        if not rect:
            scene_lsize = logical_size
            scene_psize = physical_size
            physical_viewport = 0, 0, physical_size[0], physical_size[1], 0, 1
        elif len(rect) == 4:
            scene_lsize = rect[2], rect[3]
            physical_viewport = [int(i * pixel_ratio + 0.4999) for i in rect]
            physical_viewport = (*physical_viewport, 0, 1)
            scene_psize = physical_viewport[2], physical_viewport[3]
        else:
            raise ValueError(
                "The viewport rect must be None or 4 elements (x, y, w, h)."
            )

        # Apply the camera's native size (do this before we change scene_lsize based on view_offset)
        camera.set_view_size(*scene_lsize)

        # Camera view_offset overrides logical size
        ndc_offset = (1.0, 1.0, 0.0, 0.0)  # (ax ay bx by)  virtual_ndc = a * ndc + b
        if camera._view_offset is not None:
            scene_lsize = camera._view_offset["width"], camera._view_offset["height"]
            ndc_offset = camera._view_offset["ndc_offset"]

        # Allow objects to prepare just in time. When doing multiple
        # render calls, we don't want to spam. The clear flag is
        # a good indicator to detect the first render call.
        if clear:
            ev = WindowEvent(
                "before_render",
                target=None,
                root=self,
                width=logical_size[0],
                height=logical_size[1],
                pixel_ratio=pixel_ratio,
            )
            self.dispatch_event(ev)

        # Get a flat and sorted version of the scene.
        # This is also where wobject._update_object() is called
        view_matrix = None
        if self._sort_objects:
            view_matrix = camera.view_matrix  # == camera.world.inverse_matrix
        flat = FlatScene(scene, view_matrix, self._object_count)
        self._object_count = flat.object_count
        flat.sort()

        # Prepare the shared object
        self._shared.pre_render_hook()

        # Update stdinfo uniform buffer object that we'll use during this render call
        self._update_stdinfo_buffer(camera, scene_psize, scene_lsize, ndc_offset)

        # Get renderstate object
        renderstate = get_renderstate(flat.lights, self._blender)
        self._renderstates_per_flush[0].append(renderstate)

        # Make sure pipeline objects exist for all wobjects. This also collects the bake functons.
        flat.collect_pipelines_container_groups(renderstate)

        # Enable pipelines to update data on the CPU.
        flat.call_bake_functions(camera, logical_size)

        # Update *all* buffers and textures that have changed
        for resource in resource_update_registry.get_syncable_resources(flush=True):
            update_resource(resource)

        # Command buffers cannot be reused. If we want some sort of re-use we should
        # look into render bundles. See https://github.com/gfx-rs/wgpu-native/issues/154
        # If we do get this to work, we should trigger a new recording
        # when the wobject's children, visible, render_order, or render_pass changes.

        # Record the rendering of all world objects, or re-use previous recording
        command_buffers = []
        command_buffers += self._render_recording(
            renderstate,
            flat,
            physical_viewport,
        )

        # Collect commands and submit
        self._device.queue.submit(command_buffers)

        if flush:
            self.flush()

    def flush(self, target=None):
        """Render the result into the target. This method is called
        automatically unless you use ``.render(..., flush=False)``.
        """

        # Print FPS
        now = time.perf_counter()
        if self._show_fps:
            if self._fps["count"] == 0:
                print(f"Time to first draw: {now - self._fps['start']:0.2f}")
                self._fps["start"] = now
                self._fps["count"] = 1
            elif now > self._fps["start"] + 1:
                fps = self._fps["count"] / (now - self._fps["start"])
                print(f"FPS: {fps:0.1f}")
                self._fps["start"] = now
                self._fps["count"] = 1
            else:
                self._fps["count"] += 1

        need_mipmaps = False
        if target is None:
            target = self._target

        # Get the target texture view.
        if isinstance(target, BaseRenderCanvas):
            target_tex = self._canvas_context.get_current_texture().create_view()
        elif isinstance(target, Texture):
            need_mipmaps = target.generate_mipmaps
            target_tex = getattr(target, "_wgpu_default_view", None)
            if target_tex is None:
                target_tex = ensure_wgpu_object(GfxTextureView(target))
                target._wgpu_default_view = target_tex
        elif isinstance(target, GfxTextureView):
            need_mipmaps = target.texture.generate_mipmaps
            target_tex = ensure_wgpu_object(target)
        else:
            raise TypeError("Unexpected target type.")

        # Reset counter (so we can auto-clear the first next draw)
        self._renders_since_last_flush = 0

        # Start recording ...
        command_encoder = self._device.create_command_encoder()

        # Preparations
        src_name, dst_name = "color", "altcolor"
        src_usage = wgpu.TextureUsage.TEXTURE_BINDING
        dst_usage = wgpu.TextureUsage.RENDER_ATTACHMENT

        if self._name_of_texture_with_effects:
            # probably flushing a second time
            src_name = self._name_of_texture_with_effects
        else:
            # Apply any effect passes
            for step in self._effect_passes:
                color_tex = self._blender.get_texture_view(
                    src_name, src_usage, create_if_not_exist=True
                )
                dst_tex = self._blender.get_texture_view(
                    dst_name, dst_usage, create_if_not_exist=True
                )
                depth_tex = None
                if step.USES_DEPTH:
                    depth_tex = self._blender.get_texture_view(
                        "depth", src_usage, create_if_not_exist=False
                    )
                step.render(command_encoder, color_tex, depth_tex, dst_tex)
                # Pingpong
                src_name, dst_name = dst_name, src_name
            self._name_of_texture_with_effects = src_name

        # Apply copy-pass
        color_tex = self._blender.get_texture_view(src_name, src_usage)
        self._output_pass.gamma = self._gamma_correction * self._gamma_correction_srgb
        # self._output_pass.filter_strength = self._pixel_filter
        self._output_pass.render(command_encoder, color_tex, None, target_tex)

        self._device.queue.submit([command_encoder.finish()])

        if need_mipmaps:
            generate_texture_mipmaps(target)

    def _render_recording(
        self,
        renderstate,
        flat,
        physical_viewport,
    ):
        # You might think that this is slow for large number of world
        # object. But it is actually pretty good. It does iterate over
        # all world objects, and over stuff in each object. But that's
        # it, really.
        # todo: we may be able to speed this up with render bundles though

        command_encoder = self._device.create_command_encoder()

        # ----- compute pipelines

        compute_pass = command_encoder.begin_compute_pass()
        for compute_pipeline_container in flat.iter_compute_pipelines():
            compute_pipeline_container.dispatch(compute_pass)
        compute_pass.end()

        # ----- process shadow maps

        if flat.shadow_objects:
            lights = (
                flat.lights["point_lights"]
                + flat.lights["spot_lights"]
                + flat.lights["directional_lights"]
            )
            render_shadow_maps(lights, flat.shadow_objects, command_encoder)

        # ----- render in stages

        blender = self._blender
        rendered_something = False

        for iter in flat.iter_render_pipelines_per_pass_type():
            pass_type, render_pipeline_containers = iter

            # Only render this pass type if there are objects
            if not render_pipeline_containers:
                continue
            rendered_something = True

            render_pass = command_encoder.begin_render_pass(
                color_attachments=blender.get_color_attachments(pass_type),
                depth_stencil_attachment=blender.get_depth_attachment(),
                occlusion_query_set=None,
            )

            render_pass.set_viewport(*physical_viewport)

            for render_pipeline_container in render_pipeline_containers:
                render_pipeline_container.draw(render_pass, renderstate)

            render_pass.end()

            # Objects blended with weighted must be resolved into the color texture
            if pass_type == "weighted":
                self._blender.perform_weighted_resolve_pass(command_encoder)

        # Make sure the render targets exist (even if no objects are rendered)
        if not rendered_something:
            blender.get_color_attachments("normal")
            blender.get_depth_attachment()

        return [command_encoder.finish()]

    def _update_stdinfo_buffer(
        self, camera: Camera, physical_size, logical_size, ndc_offset
    ):
        # Update the stdinfo buffer's data
        # All matrices need to be transposed, because in WGSL they are column-major.
        # -> From the numpy p.o.v. the matrices are transposed, in wgsl they are
        #    upright again because the different interpretation of the memory.
        stdinfo_data = self._shared.uniform_buffer.data
        stdinfo_data["cam_transform"] = camera.world.inverse_matrix.T
        stdinfo_data["cam_transform_inv"] = camera.world.matrix.T
        stdinfo_data["projection_transform"] = camera.projection_matrix.T
        stdinfo_data["projection_transform_inv"] = camera.projection_matrix_inverse.T
        # stdinfo_data["ndc_to_world"].flat = la.mat_inverse(stdinfo_data["cam_transform"] @ stdinfo_data["projection_transform"])
        stdinfo_data["ndc_offset"] = ndc_offset
        stdinfo_data["physical_size"] = physical_size
        stdinfo_data["logical_size"] = logical_size
        # Upload to GPU
        self._shared.uniform_buffer.update_full()

    # Picking

    def get_pick_info(self, pos):
        """Get information about the given window location. The given
        pos is a 2D point in logical pixels (with the origin at the
        top-left). Returns a dict with fields:

        * "ndc": The position in normalized device coordinates, the 3d element
            being the depth (0..1). Can be translated to the position
            in world coordinates using the camera transforms.
        * "rgba": The value in the color buffer. All zero's when rendering
          directly to the screen (bypassing post-processing).
        * "world_object": the object at that location (provided that
          the object supports picking).
        * Additional pick info may be available, depending on the type of
          object and its material. See the world-object classes for details.
        """

        # Make pos 0..1, so we can scale it to the render texture
        logical_size = self.logical_size
        float_pos = pos[0] / logical_size[0], pos[1] / logical_size[1]

        # Prevent out of range and picking before first draw
        out_of_range = not (0 <= float_pos[0] <= 1 and 0 <= float_pos[1] <= 1)
        blender_zero_size = self._blender.size == (0, 0)
        if out_of_range or blender_zero_size:
            return {"rgba": Color(0, 0, 0, 0), "world_object": None}

        color_tex = self._blender.get_texture("color")
        pick_tex = self._blender.get_texture("pick")

        # Sample
        encoder = self._device.create_command_encoder()
        if color_tex:
            self._copy_pixel(encoder, color_tex, float_pos, 0)
        if pick_tex:
            self._copy_pixel(encoder, pick_tex, float_pos, 8)
        self._device.queue.submit([encoder.finish()])

        # Collect data from the buffer
        self._pixel_info_buffer.map_sync("read")
        try:
            data = self._pixel_info_buffer.read_mapped()
        finally:
            self._pixel_info_buffer.unmap()
        color = Color(x / 255 for x in tuple(data[0:4].cast("B")))
        pick_value = tuple(data[8:16].cast("Q"))[0]  # noqa: RUF015
        wobject_id = pick_value & 1048575  # 2**20-1
        wobject = id_provider.get_object_from_id(wobject_id)
        # Note: the position in world coordinates is not included because
        # it depends on the camera, but we don't "own" the camera.

        info = {
            "rgba": color,
            "world_object": wobject,
        }

        if wobject:
            pick_info = wobject._wgpu_get_pick_info(pick_value)
            info.update(pick_info)
        return info

    def _copy_pixel(self, encoder, render_texture, float_pos, buf_offset):
        # Map position to the texture index
        w, h, _d = render_texture.size
        x = max(0, min(w - 1, int(float_pos[0] * w)))
        y = max(0, min(h - 1, int(float_pos[1] * h)))

        # Note: bytes_per_row must be a multiple of 256.
        encoder.copy_texture_to_buffer(
            {
                "texture": render_texture,
                "mip_level": 0,
                "origin": (x, y, 0),
            },
            {
                "buffer": self._pixel_info_buffer,
                "offset": buf_offset,
                "bytes_per_row": 256,  # render_texture.bytes_per_pixel,
                "rows_per_image": 1,
            },
            copy_size=(1, 1, 1),
        )

    def snapshot(self):
        """Create a snapshot of the currently rendered image."""

        # Prepare
        texture = self._blender.get_texture(
            self._name_of_texture_with_effects or "color"
        )
        size = texture.size
        if texture.format == "rgba16float":
            bytes_per_pixel = 8
            dtype = np.float16
        else:
            bytes_per_pixel = 4
            dtype = np.uint8

        # Note, with queue.read_texture the bytes_per_row limitation does not apply.
        data = self._device.queue.read_texture(
            {
                "texture": texture,
                "mip_level": 0,
                "origin": (0, 0, 0),
            },
            {
                "offset": 0,
                "bytes_per_row": bytes_per_pixel * size[0],
                "rows_per_image": size[1],
            },
            size,
        )

        return np.frombuffer(data, dtype).reshape(size[1], size[0], 4)

    def request_draw(self, draw_function=None):
        """Forwards a request_draw call to the target canvas. If the renderer's
        target is not a canvas (e.g. a texture) this function does
        nothing.
        """
        request_draw = getattr(self.target, "request_draw", None)
        if request_draw:
            request_draw(draw_function)

    def enable_events(self):
        """Add event handlers for a specific list of events that are generated
        by the canvas. The handler is the ``convert_event`` method in order to
        convert the Wgpu event dicts into Pygfx event objects."""

        # Check for ``add_event_handler`` attribute. Silently 'fail' as it is
        # perfectly fine to pass a texture as a target to the renderer
        if hasattr(self._target, "add_event_handler"):
            self._target.add_event_handler(self.convert_event, *EVENTS_TO_CONVERT)

    def disable_events(self):
        """Remove the event handlers from the canvas."""
        if hasattr(self._target, "remove_event_handler"):
            self._target.remove_event_handler(self.convert_event, *EVENTS_TO_CONVERT)

    def convert_event(self, event: dict):
        """Converts Wgpu event (dict following jupyter_rfb spec) to Pygfx Event object,
        adds picking info and then dispatches the event in the Pygfx event system.
        """
        event_type = event["event_type"]
        target = None
        if "x" in event and "y" in event:
            info = self.get_pick_info((event["x"], event["y"]))
            target = info["world_object"]
            event["pick_info"] = info

        ev = EVENT_TYPE_MAP[event_type](
            type=event_type, **event, target=target, root=self
        )
        self.dispatch_event(ev)


EVENT_TYPE_MAP = {
    "resize": WindowEvent,
    "close": WindowEvent,
    "pointer_down": PointerEvent,
    "pointer_up": PointerEvent,
    "pointer_move": PointerEvent,
    "double_click": PointerEvent,
    "wheel": WheelEvent,
    "key_down": KeyboardEvent,
    "key_up": KeyboardEvent,
}


EVENTS_TO_CONVERT = (
    "key_down",
    "key_up",
    "pointer_down",
    "pointer_move",
    "pointer_up",
    "wheel",
    "close",
    "resize",
)
